
#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>

#include "google/profiler.h"

#include "gdsl/gdsl_perm.h"
#include "gdsl/gdsl_types.h"
#include "gdsl/gdsl_rbtree.h"

struct timeval start;
struct timeval end;
float  time_accum = 0;

static int profilingOn = 0;
static pthread_mutex_t mutex;

FILE *trace;

void __POKE__ProfilingBegin() {

  const char *pprof = getenv("PPROF");

  if (pprof) {
    fprintf(stderr, "__POKE__ Google profiling turned ON\n");
    ProfilerStart(pprof);
  } else {
    fprintf(stderr, "__POKE__ LLVM profiling turned ON\n");
    profilingOn = 1;
  }

  fprintf(stderr, "__POKE__ Timing turned ON\n");
  gettimeofday(&start, 0);
}

void __POKE__ProfilingEnd() {

  if (getenv("PPROF")) {
    fprintf(stderr, "__POKE__ Google profiling turned OFF\n");
    ProfilerStop();
  } else {
    fprintf(stderr, "__POKE__ LLVM profiling turned OFF\n");
    profilingOn = 0;
  }

  gettimeofday(&end, 0);
  time_accum += ((float)(end.tv_sec  - start.tv_sec) +
                 (float)(end.tv_usec - start.tv_usec) / 1000000.0);

  fprintf(stderr, "__POKE__ Time: %f\n", time_accum);
}

void __POKE__ProfilingAuto() {
  __POKE__ProfilingBegin();
  atexit(__POKE__ProfilingEnd);
}


/* backward compatibility */
void poke_begin_instrumentation() { __POKE__ProfilingBegin(); }
void poke_end_instrumentation()   { __POKE__ProfilingEnd(); }


/* Profiling extensions
 * -------------------- */

static unsigned long long *counterArray;

int llvm_start_block_profiling(int argc, const char **argv,
        unsigned long long *arrayStart, unsigned numElements);

int __POKE__ProfilingInit(int argc, const char **argv, unsigned long
        long *array, unsigned numElements)
{
  pthread_mutex_init(&mutex, NULL);

  counterArray = array;
  return llvm_start_block_profiling(argc, argv, array, numElements);
}

void __POKE__ProfileAccum(unsigned functionIndex,
                          unsigned *array,
                          unsigned arraySize)
{
  if (profilingOn) {
    unsigned i;
    pthread_mutex_lock(&mutex);

    for (i = 0; i < arraySize; ++i)
      counterArray[functionIndex + i] += array[i];

    pthread_mutex_unlock(&mutex);
  }
}

/* InstCounter extensions
 * ------------------ */

static unsigned long long __POKE__InstCounter = 0;
static unsigned long long __POKE__RealCounter = 0;
static unsigned long long __POKE__RealRecCounter = 0;

void __POKE__InstCounterExit() {
  fprintf(stderr, "__POKE__ InstCount: %llu\n", __POKE__InstCounter);
  fprintf(stderr, "__POKE__ RealCount: %llu\n", __POKE__RealCounter);
  fprintf(stderr, "__POKE__ RealRecCount: %llu\n", __POKE__RealRecCounter);
}

void __POKE__InstCounterInit()
{
  pthread_mutex_init(&mutex, NULL);
  atexit(__POKE__InstCounterExit);
}

void __POKE__InstCounterAccum(unsigned instCount)
{
  if (profilingOn) {

    pthread_mutex_lock(&mutex);
    __POKE__InstCounter += instCount;
    pthread_mutex_unlock(&mutex);
  }
}

void __POKE__RegisterReal() {
  pthread_mutex_lock(&mutex);
  __POKE__RealCounter++;
  pthread_mutex_unlock(&mutex);
}

void __POKE__RegisterRealRecover() {
  pthread_mutex_lock(&mutex);
  __POKE__RealRecCounter++;
  pthread_mutex_unlock(&mutex);
}

/* Relaxer */

gdsl_rbtree_t t;

struct RelaxPair {
  int key;
  long long val;
};

gdsl_element_t alloc_pair(void *in_pair) {
  struct RelaxPair *out_pair = malloc(sizeof(struct RelaxPair));
  struct RelaxPair *in_pair2 = in_pair;
  out_pair->key = in_pair2->key;
  out_pair->val = in_pair2->val;
  return (gdsl_element_t) out_pair;
}

void free_pair(gdsl_element_t e) {
  free(e);
}

long int compare_pairs(gdsl_element_t p1, void *p2) {
  int left  = ((struct RelaxPair *)p1)->key;
  int right = ((struct RelaxPair *)p2)->key;

  return (left - right);
}

void print_pair(gdsl_element_t e, FILE *file, gdsl_location_t location, void *d) {
  struct RelaxPair *pair = (struct RelaxPair *)e;
  fprintf(file, "%d %lld\n", pair->key, pair->val);
}


void __POKE__RelaxDestroy() {
  trace = fopen("trace.out", "w");
  gdsl_rbtree_write(t, print_pair, trace, NULL);
  fclose(trace);
}

void __POKE__RelaxInit() {

  t = gdsl_rbtree_alloc("relaxer_map", alloc_pair, free_pair, compare_pairs);
  atexit(__POKE__RelaxDestroy);
}

static void IncrementPair(int instCount) {
  int rc;

  struct RelaxPair pair;
  struct RelaxPair *inserted_pair;

  pair.key = instCount;
  pair.val = 0;

  inserted_pair = (struct RelaxPair *)gdsl_rbtree_insert(t, (void *)&pair, &rc);

  inserted_pair->val++;
}

void __POKE__RelaxTrans(int instCount) {
  IncrementPair(instCount);
}

void __POKE__RelaxOff(int instCount) {
  IncrementPair(instCount);
  IncrementPair(-1);
}


