
#include "llvm/Constants.h"
#include "llvm/Function.h"
#include "llvm/InstrTypes.h"
#include "llvm/Instruction.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Use.h"
#include "llvm/Support/Streams.h"
#include "llvm/Support/Debug.h"

#include "FaultInjectionUtils.h"
#include "InstructionInjection.h"
#include "Snapshot.h"

#include <map>

using namespace llvm;
using namespace vertical;


namespace {
  class AssistedInjection : public InstructionInjection {
    bool runOnFunction(Function &F);
    void getAnalysisUsage(AnalysisUsage &AU) const;

    bool isInstrumented(Instruction *I);

  public:
    static char ID;
    AssistedInjection() : InstructionInjection((intptr_t)&ID) {}
  };
}


char AssistedInjection::ID = 0;
static RegisterPass<AssistedInjection> X(
    "assisted", "Insert assisted fault-injection semantics");


void AssistedInjection::getAnalysisUsage(AnalysisUsage &AU) const {
  // AU.addPreserved<PoisonExtend>();
  // if (UsePoison)
  //   AU.addRequired<PoisonExtend>();
  AU.addRequired<Snapshot>();
  InstructionInjection::getAnalysisUsage(AU);
}


bool AssistedInjection::runOnFunction(Function &F) {
  
  Snapshot &SS  = getAnalysis<Snapshot>();

  forkInsts(F);

  for (std::map<Instruction *, Instruction *>::iterator IMI =
       InjectionMap.begin(); IMI != InjectionMap.end(); ++IMI) {

    Instruction *OrigI = IMI->first;
    Instruction *FailI = IMI->second;

    DOUT << "AssistedInjection: " << SS.IIMap[OrigI]->StringId << "\n";

    // BasicBlock *OrigBlock = OrigI->getParent();
    BasicBlock *FailBlock = FailI->getParent();

    assert(FailI->use_empty() || FailI->hasOneUse());
    Use &UsePoint = FailI->use_begin().getUse();

    Value *UseReplacement = 0;

    switch (FailI->getOpcode()) {

      case Instruction::GetElementPtr: {
        // return first operand pointer 
        GetElementPtrInst *GetElementPtr = cast<GetElementPtrInst>(FailI);
        UseReplacement = new BitCastInst(
            GetElementPtr->getPointerOperand(),
            GetElementPtr->getType(), "PtrOnly",
            FailBlock->getTerminator());
        break;
      }

      case Instruction::Load:
        // load zero 
        UseReplacement = Constant::getNullValue(
            cast<PointerType>(
                cast<LoadInst>(FailI)->getPointerOperand()->getType()
                )->getElementType());

        break;


      case Instruction::Store:

        // silent failure
        FailI->eraseFromParent();
        InjectionMap[OrigI] = FailBlock->getTerminator();

        // if (UsePoison) {
        //   // mark the store poison bit
        //   GlobalVariable *StorePoisonBit =
        //     M.getGlobalVariable("__POKE__StorePoisonBit", true);
        //   new StoreInst(PoisonExtend::On, StorePoisonBit, FailI);
        // }
        break;


      case Instruction::ICmp:
      case Instruction::FCmp:
        // invert the failed instruction output 
        UseReplacement = BinaryOperator::Create(
            Instruction::Xor, FailI,
            ConstantInt::get(Type::Int1Ty, 1),
            "SwapCmp",
            FailBlock->getTerminator());
        break;


      case Instruction::Select: {
        SelectInst *Select = cast<SelectInst>(FailI);

        // swap sources 
        Value *TempValue = Select->getOperand(1);
        Select->setOperand(1, Select->getOperand(2));
        Select->setOperand(2, TempValue);
        break;
      }


      case Instruction::Br: {

        BranchInst *Branch = cast<BranchInst>(FailI);

        if (!Branch->isConditional())
          break;

        // TODO: fix this hack
        //if (ControlAbort) {

        //  CreateNullReturn(Branch);

        //  // Update the PHI nodes
        //  for (unsigned i = 0; i < Branch->getNumSuccessors(); ++i) {
        //    BasicBlock *Successor = Branch->getSuccessor(i);

        //    PHINode *PN;
        //    for (BasicBlock::iterator I = Successor->begin();
        //         (PN = dyn_cast<PHINode>(I)); ++I) {

        //      int IDX = PN->getBasicBlockIndex(Branch->getParent());
        //      while (IDX != -1) {
        //        PN->removeIncomingValue(IDX);
        //        IDX = PN->getBasicBlockIndex(Branch->getParent());
        //      }
        //    }
        //  }

        //  EraseI = true;

        //} else {

        // if (UsePoison) {
        //   // mark the branch poison bit
        //   GlobalVariable *BranchPoisonBit =
        //     M.getGlobalVariable("__POKE__BranchPoisonBit", true);
        //   new StoreInst(PoisonExtend::On, BranchPoisonBit, Branch);
        // } 

        // swap destinations 
        BasicBlock *TempSucc = Branch->getSuccessor(0);
        Branch->setSuccessor(0, Branch->getSuccessor(1));
        Branch->setSuccessor(1, TempSucc);

        break;
      }


      case Instruction::Switch: {

        // if (UsePoison) {
        //   // Mark the branch poison bit
        //   GlobalVariable *BranchPoisonBit =
        //     M.getGlobalVariable("__POKE__BranchPoisonBit", true);
        //   new StoreInst(PoisonExtend::On, BranchPoisonBit, Switch);
        // }

        // Branch to the default destination
        SwitchInst *Switch = cast<SwitchInst>(FailI);
        BranchInst *Replacement = BranchInst::Create(Switch->getDefaultDest(),
                                                     Switch);

        // Update the PHI nodes for all successors of FailBlock that are now no
        // longer considered 
        for (unsigned i = 0; i < Switch->getNumSuccessors(); ++i) {
          BasicBlock *Successor = Switch->getSuccessor(i);

          if (Successor == Switch->getDefaultDest())
            // this one doesn't change
            continue;

          PHINode *PN;
          for (BasicBlock::iterator I = Successor->begin();
               (PN = dyn_cast<PHINode>(I)); ++I) {

            int IDX = PN->getBasicBlockIndex(Switch->getParent());
            while (IDX != -1) {
              PN->removeIncomingValue(IDX);
              IDX = PN->getBasicBlockIndex(Switch->getParent());
            }
          }
        }

        FailI->eraseFromParent();
        InjectionMap[OrigI] = Replacement;
        break;
      }


      case Instruction::Add:
      case Instruction::Sub:
      case Instruction::Mul:
      case Instruction::UDiv:
      case Instruction::SDiv:
      case Instruction::FDiv:
      case Instruction::URem:
      case Instruction::SRem:
      case Instruction::FRem:
      case Instruction::Shl:
      case Instruction::LShr:
      case Instruction::AShr:
      case Instruction::And:
      case Instruction::Or:
      case Instruction::Xor:

        // take first operand 
        UseReplacement = cast<BinaryOperator>(FailI)->getOperand(0);
        break;


      case Instruction::FPToUI:
      case Instruction::FPToSI:
      case Instruction::UIToFP:
      case Instruction::SIToFP: {

        // set to zero 
        CastInst *Cast = cast<CastInst>(FailI);
        Cast->setOperand(
            0, Constant::getNullValue(Cast->getOperand(0)->getType()));
        break;
      }


      default:
        assert(0 && "Unsupported opcode");
        break;
    }

    if (UseReplacement) {
      UsePoint.set(UseReplacement);
    }

    // if (UsePoison && !OrigI->use_empty()) {

    //   Value *PoisonBit = getAnalysis<PoisonExtend>().PoisonBitMap[OrigI];
    //   assert(PoisonBit != 0);

    //   assert(OrigBlock->getTerminator()->getNumSuccessors() == 1);
    //   assert(FailBlock->getTerminator()->getNumSuccessors() == 1);
    //   assert(OrigBlock->getTerminator()->getSuccessor(0) ==
    //          FailBlock->getTerminator()->getSuccessor(0));

    //   BasicBlock *NextBlock = OrigBlock->getTerminator()->getSuccessor(0);

    //   // Insert a PHINode before the successor block
    //   PHINode *PN = PHINode::Create(
    //       PoisonBit->getType(), "IncomingPoison",
    //       NextBlock->begin());

    //   // Replacing uses of PoisonBit with PHI value
    //   PoisonBit->replaceAllUsesWith(PN);

    //   PN->addIncoming(PoisonBit, OrigBlock);
    //   PN->addIncoming(PoisonExtend::On, FailBlock);
    // }
  }

  return true;
}


bool AssistedInjection::isInstrumented(Instruction *I)
{
  switch (I->getOpcode()) {

    // ptr computation/access
    case Instruction::GetElementPtr:
    case Instruction::Load:
    case Instruction::Store:
    // comparison
    case Instruction::ICmp:
    case Instruction::FCmp:
    // control
    case Instruction::Select:
    case Instruction::Br:
    case Instruction::Switch:
    // arithmetic/logic
    case Instruction::Add:
    case Instruction::Sub:
    case Instruction::Mul:
    case Instruction::UDiv:
    case Instruction::SDiv:
    case Instruction::FDiv:
    case Instruction::URem:
    case Instruction::SRem:
    case Instruction::FRem:
    case Instruction::Shl:
    case Instruction::LShr:
    case Instruction::AShr:
    case Instruction::And:
    case Instruction::Or:
    case Instruction::Xor:
    // casts
    case Instruction::FPToUI:
    case Instruction::FPToSI:
    case Instruction::UIToFP:
    case Instruction::SIToFP:
      // supported
      return true;

    default:
      // unsupported
      return false;
  }
}

