
#include "llvm/Constants.h"
#include "llvm/Function.h"
#include "llvm/InstrTypes.h"
#include "llvm/Instruction.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Use.h"
#include "llvm/Support/Streams.h"
#include "llvm/Support/Debug.h"

#include "FaultInjectionUtils.h"
#include "InstructionInjection.h"
#include "Snapshot.h"

#include <map>

using namespace llvm;
using namespace vertical;


namespace {
  class BitFlipInjection : public InstructionInjection {
    bool runOnFunction(Function &F);
    void getAnalysisUsage(AnalysisUsage &AU) const;

    bool isInstrumented(Instruction *I);

  public:
    static char ID;
    BitFlipInjection() : InstructionInjection((intptr_t)&ID) {}
  };
}


char BitFlipInjection::ID = 0;
static RegisterPass<BitFlipInjection> X(
    "bitflip", "Insert bitflip fault-injection semantics");


void BitFlipInjection::getAnalysisUsage(AnalysisUsage &AU) const {
  // AU.addPreserved<PoisonExtend>();
  // if (UsePoison)
  //   AU.addRequired<PoisonExtend>();
  AU.addRequired<Snapshot>();
  InstructionInjection::getAnalysisUsage(AU);
}


static Value *Flip(Value *Val, Instruction *InsertPos) 
{
  Module *M = InsertPos->getParent()->getParent()->getParent();
  const Type *Ty  = Val->getType();
  const Type *FlipTy;

  bool                 IsCastNecessary;
  Instruction::CastOps ToCastOp;
  Instruction::CastOps FromCastOp;

  if (isa<PointerType>(Ty)) {
    FlipTy          = Type::Int64Ty;
    ToCastOp        = Instruction::PtrToInt;
    FromCastOp      = Instruction::IntToPtr;
    IsCastNecessary = true;

  } else if (Ty->isFloatingPoint()) {
    FlipTy          = IntegerType::get(Ty->getPrimitiveSizeInBits());
    ToCastOp        = Instruction::BitCast;
    FromCastOp      = Instruction::BitCast;
    IsCastNecessary = true;

  } else {
    FlipTy          = Ty;
    IsCastNecessary = false;
    assert(Ty->isInteger() && "unsupported type");
  }

  unsigned FlipTySize = FlipTy->getPrimitiveSizeInBits();
  Value *XorMask;

  if (FlipTySize == 1) {
    DOUT << "Creating XOR mask\n";
    XorMask = ConstantInt::get(FlipTy, 1);

  } else {
    DOUT << "Generating call to function '__POKE__RandUIntMasked'\n";
    Constant *RandUIntFunc =
      M->getOrInsertFunction("__POKE__RandUIntMasked",
                             Type::Int32Ty, Type::Int32Ty, (Type *)0);

    Value *WhichBit = CallInst::Create(
        RandUIntFunc,
        ConstantInt::get(Type::Int32Ty, FlipTySize - 1),
        "whichbit",
        InsertPos);

    if (FlipTySize != 32) {
      DOUT << "Casting bitshift to integer type\n";

      Instruction::CastOps CastOp;
      if (FlipTySize > 32)
        CastOp = Instruction::SExt;
      else
        CastOp = Instruction::Trunc;

      WhichBit = CastInst::Create(
          CastOp,
          WhichBit,
          FlipTy,
          "whichbitcast",
          InsertPos);
    }

    DOUT << "Creating XOR mask\n";
    XorMask = BinaryOperator::Create(
        Instruction::Shl,
        ConstantInt::get(FlipTy, 1),
        WhichBit,
        "xormask",
        InsertPos);
  }

  Value *FlippedVal = Val;

  if (IsCastNecessary) {
    DOUT << "Casting instruction to integer type\n";
    FlippedVal = CastInst::Create(
        ToCastOp,
        FlippedVal,
        FlipTy,
        "tointegercast",
        InsertPos);
  }

  DOUT << "Creating XOR instruction\n";
  FlippedVal = BinaryOperator::Create(
      Instruction::Xor,
      FlippedVal,
      XorMask,
      "bitflip",
      InsertPos);

  if (IsCastNecessary) {
    DOUT << "Casting back to original type\n";
    FlippedVal = CastInst::Create(
        FromCastOp,
        FlippedVal,
        Ty,
        "fromintegercast",
        InsertPos);
  }

  return FlippedVal;
}


bool BitFlipInjection::runOnFunction(Function &F) {
  
  Snapshot &SS  = getAnalysis<Snapshot>();

  forkInsts(F);

  for (std::map<Instruction *, Instruction *>::iterator IMI =
       InjectionMap.begin(); IMI != InjectionMap.end(); ++IMI) {

    Instruction *OrigI = IMI->first;
    Instruction *FailI = IMI->second;

    DOUT << "BitFlipInjection: " << SS.IIMap[OrigI]->StringId << "\n";

    // BasicBlock *OrigBlock = OrigI->getParent();
    BasicBlock *FailBlock = FailI->getParent();

    switch (FailI->getOpcode()) {

      // -- Stores:
      // stores are special
      case Instruction::Store: {

        Value *Val = FailI->getOperand(0);
        Value *FlippedVal = Flip(Val, FailI);

        FailI->setOperand(0, FlippedVal);
        break;
      } 

      default: {

        if (FailI->hasOneUse()) {
          Use &UsePoint = FailI->use_begin().getUse();

          Value *FlippedVal = Flip(FailI, FailBlock->getTerminator());
          UsePoint.set(FlippedVal);
          break;

        } else {
          assert(FailI->use_empty());
        }
      }
    }

    // if (UsePoison && !OrigI->use_empty()) {

    //   Value *PoisonBit = getAnalysis<PoisonExtend>().PoisonBitMap[OrigI];
    //   assert(PoisonBit != 0);

    //   assert(OrigBlock->getTerminator()->getNumSuccessors() == 1);
    //   assert(FailBlock->getTerminator()->getNumSuccessors() == 1);
    //   assert(OrigBlock->getTerminator()->getSuccessor(0) ==
    //          FailBlock->getTerminator()->getSuccessor(0));

    //   BasicBlock *NextBlock = OrigBlock->getTerminator()->getSuccessor(0);

    //   // Insert a PHINode before the successor block
    //   PHINode *PN = PHINode::Create(
    //       PoisonBit->getType(), "IncomingPoison",
    //       NextBlock->begin());

    //   // Replacing uses of PoisonBit with PHI value
    //   PoisonBit->replaceAllUsesWith(PN);

    //   PN->addIncoming(PoisonBit, OrigBlock);
    //   PN->addIncoming(PoisonExtend::On, FailBlock);
    // }
  }

  return true;
}


bool BitFlipInjection::isInstrumented(Instruction *I)
{
  switch (I->getOpcode()) {
    // memory creation (ADDED!)
    case Instruction::Malloc:
    // call instructions (ADDED!)
    case Instruction::Call:
    // ptr computation/access
    case Instruction::GetElementPtr:
    case Instruction::Load:
    case Instruction::Store:
    // comparison
    case Instruction::ICmp:
    case Instruction::FCmp:
    // control
    case Instruction::Select:
    // arithmetic/logic
    case Instruction::Add:
    case Instruction::Sub:
    case Instruction::Mul:
    case Instruction::UDiv:
    case Instruction::SDiv:
    case Instruction::FDiv:
    case Instruction::URem:
    case Instruction::SRem:
    case Instruction::FRem:
    case Instruction::Shl:
    case Instruction::LShr:
    case Instruction::AShr:
    case Instruction::And:
    case Instruction::Or:
    case Instruction::Xor:
    // casts
    case Instruction::FPToUI:
    case Instruction::FPToSI:
    case Instruction::UIToFP:
    case Instruction::SIToFP:
    // casts (ADDED!)
    case Instruction::Trunc:
    case Instruction::ZExt:
    case Instruction::SExt:
    case Instruction::FPTrunc:
    case Instruction::FPExt:
    case Instruction::PtrToInt:
    case Instruction::IntToPtr:
    case Instruction::BitCast:
      // supported
      return true;

    default:
      // unsupported
      return false;
  }
}

