
#include "llvm/Function.h"
#include "llvm/Instruction.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Use.h"
#include "llvm/ADT/hash_map.h"
#include "llvm/Support/Streams.h"
#include "llvm/Support/Debug.h"

#include <iomanip>

#include "FaultInjectionUtils.h"
#include "Criticality.h"
#include "Snapshot.h"

using namespace llvm;
using namespace vertical;


char Criticality::ID = 0;
static RegisterPass<Criticality> X(
    "criticality", "Criticality Analysis", true, true);

void Criticality::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequiredTransitive<Snapshot>();
  AU.setPreservesAll();
}

bool Criticality::runOnFunction(Function &F) {

  CriticalityMap.clear();
  LastFn = &F;

  Snapshot &SS = getAnalysis<Snapshot>();
  for (Snapshot::const_ii_iterator II = SS.ii_begin(), IE = SS.ii_end();
       II != IE; ++II) {
    CriticalityMap[(*II)->I] = 0;
  }

  for (Snapshot::const_ii_iterator II = SS.ii_begin(), IE = SS.ii_end();
       II != IE; ++II) {
    Instruction *I = (*II)->I;

    uint32_t     bitMask = 0;
    Instruction *Source  = 0;

    if (BranchInst *B = dyn_cast<BranchInst>(I)) {
      Source = B;
      bitMask = 1 << 0;

    } else if (LoadInst *L = dyn_cast<LoadInst>(I)) {
      if (Instruction *Op = dyn_cast<Instruction>(L->getPointerOperand())) {
        Source = Op;
        bitMask = 1 << 1;
    } 

    } else if (StoreInst *S = dyn_cast<StoreInst>(I)) {
      if (Instruction *Op = dyn_cast<Instruction>(S->getPointerOperand())) {
        Source = Op;
        bitMask = 1 << 2;
      }
    }


    if (bitMask) {
      assert(Source);
      markCritical(Source, bitMask, true);
    }
  }

  return false;
}
    

void Criticality::markCritical(Instruction *I, uint32_t bitMask, bool first) {

  if (first)
    CriticalityMap[I] |= (bitMask << 8);

  CriticalityMap[I] |= bitMask;

  for (User::op_iterator Op = I->op_begin(); Op != I->op_end(); ++Op) {
    if (Instruction *Dep = dyn_cast<Instruction>(Op->get())) {
      if ((CriticalityMap[Dep] & bitMask) == 0) {
        // This instruction is not already marked critical; mark it and its
        // dependencies critical
        markCritical(Dep, bitMask, false);
      }
    }
  }
}


void Criticality::print(std::ostream &O, const Module *M) const {

  Snapshot &SS = getAnalysis<Snapshot>();
  for (Snapshot::const_ii_iterator II = SS.ii_begin(), IE = SS.ii_end();
       II != IE; ++II) {
    // "find" is necessary to respect constness of function
    O << CriticalityMap.find((*II)->I)->second << "\n";
  }
}


