
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Instructions.h"
#include "llvm/Constants.h"
#include "llvm/DerivedTypes.h"
#include "llvm/Function.h"
#include "llvm/Support/Debug.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"

#include "FaultInjectionUtils.h"
#include "Snapshot.h"

using namespace llvm;
using namespace vertical;


namespace {
  class DiscardInjection : public FunctionPass {
    bool runOnFunction(Function &F);
    bool doInitialization(Module &M);
    void getAnalysisUsage(AnalysisUsage &AU) const;

  public:
    static char ID;
    DiscardInjection() : FunctionPass((intptr_t)&ID) {}

    GlobalVariable *RCArray;
    GlobalVariable *ProbArray;
  };
}


char DiscardInjection::ID = 0;
static RegisterPass<DiscardInjection> X(
    "discard", "Insert function-discard fault-injection semantics");


void DiscardInjection::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequired<Snapshot>();
  AU.setPreservesAll();
}


bool DiscardInjection::doInitialization(Module &M) {

  // Initialize the global array
  const Type *FloatATy = ArrayType::get(Type::FloatTy, M.size());
  ProbArray = new GlobalVariable(
      FloatATy, false, GlobalValue::InternalLinkage,
      Constant::getNullValue(FloatATy),
      "__POKE__ProbArray", &M);

  const Type *IntATy = ArrayType::get(Type::Int32Ty, M.size());
  RCArray = new GlobalVariable(
      IntATy, false, GlobalValue::InternalLinkage,
      Constant::getNullValue(IntATy),
      "__POKE__RCArray", &M);

  // Insert the initialization functions
  const PointerType *FloatPtr = PointerType::getUnqual(Type::FloatTy);
  const PointerType *UIntPtr  = PointerType::getUnqual(Type::Int32Ty);
  Constant *PokeInitFn = M.getOrInsertFunction("__POKE__Init", Type::VoidTy,
                                               (Type *)0);
  Constant *DiscardInitFn = M.getOrInsertFunction("__POKE__DiscardInit",
                                                  Type::VoidTy, FloatPtr,
                                                  UIntPtr, Type::Int32Ty,
                                                  (Type *)0);

  Function *MainFn = M.getFunction("main");
  BasicBlock::iterator I = MainFn->begin()->begin();

  // Skip over any allocas in main's entry block.
  while (isa<AllocaInst>(I)) ++I;


  // call the poke initialization call
  CallInst::Create(PokeInitFn, "", I);

  // set the arguments for the discard initialization call and call it
  std::vector<Value*> Args(3);
  std::vector<Constant*> GEPIndices(2, Constant::getNullValue(Type::Int32Ty));
  Args[0] = ConstantExpr::getGetElementPtr(ProbArray, &GEPIndices[0],
                                           GEPIndices.size());
  Args[1] = ConstantExpr::getGetElementPtr(RCArray, &GEPIndices[0],
                                           GEPIndices.size());
  Args[2] = ConstantInt::get(Type::Int32Ty, M.size());
  CallInst::Create(DiscardInitFn, Args.begin(), Args.end(), "", I);

  return true;
}


bool DiscardInjection::runOnFunction(Function &F) {

  Snapshot &SS = getAnalysis<Snapshot>();

  DOUT << "DiscardInjection: " << F.getName() << "\n";

  // Skip over any Alloca's (no PHI's because start of function)
  BasicBlock::iterator I = F.begin()->begin();
  while (isa<AllocaInst>(I)) ++I;

  /*
   * The parent basic block of I is restructured as follows:
   *
   *      PrevBlock
   *       /    \
   *   TestBlock \
   *     /    \   \
   * FailBlock NextBlock
   *      
   *  PrevBlock contains all the logic leading up to I
   *  TestBlock contains code to check whether we should bother to generate
   *    a random number (i.e. checks whether the indexed value is non-zero)
   *  FailBlock contains the semantics for failure
   *  NextBlock contains all the logic after I in the original basic block.
   *    (If I was a terminating instruction (e.g. a branch), NextBlock does
   *    not exist.)
   */

  BasicBlock *PrevBlock = I->getParent(); 
  BasicBlock *TestBlock; 
  BasicBlock *FailBlock;
  BasicBlock *NextBlock = PrevBlock->splitBasicBlock(
      I, PrevBlock->getName() + ".Next");

  // Extending PrevBlock
  PrevBlock->getTerminator()->eraseFromParent();

  std::vector<Constant*> ProbIndices(2);
  ProbIndices[0] = Constant::getNullValue(Type::Int32Ty);
  ProbIndices[1] = ConstantInt::get(Type::Int32Ty, SS.FI->Id);
  Constant *ProbElementPtr = 
    ConstantExpr::getGetElementPtr(ProbArray, &ProbIndices[0],
                                   ProbIndices.size());
  Value *Check = new LoadInst(ProbElementPtr, "Check", PrevBlock);

  FCmpInst *IsFailable = new FCmpInst(
      FCmpInst::FCMP_OLT, Constant::getNullValue(Type::FloatTy), Check,
      "IsFailable", PrevBlock);

  // Creating TestBlock
  TestBlock = BasicBlock::Create(PrevBlock->getName() + ".Test",
                                 NextBlock->getParent(),
                                 NextBlock);

  Module *M = F.getParent();
  Constant *RandFloat = M->getOrInsertFunction("__POKE__RandFloat",
                                              Type::FloatTy, (Type *)0);

  CallInst *RandFloatCall = CallInst::Create(RandFloat, "Rand", TestBlock);

  FCmpInst *IsFailed = new FCmpInst(FCmpInst::FCMP_OLT, RandFloatCall,
                                    Check, "IsFailed", TestBlock);

  // Creating FailBlock
  FailBlock = BasicBlock::Create(PrevBlock->getName() + ".Fail",
                                 NextBlock->getParent(),
                                 NextBlock);
  std::vector<Constant*> RCIndices(2);
  RCIndices[0] = Constant::getNullValue(Type::Int32Ty);
  RCIndices[1] = ConstantInt::get(Type::Int32Ty, SS.FI->Id);
  Constant *RCElementPtr = 
    ConstantExpr::getGetElementPtr(RCArray, &RCIndices[0], RCIndices.size());
  Value *Int32RetVal = new LoadInst(RCElementPtr, "RetVal", FailBlock);

  const Type *RetType = FailBlock->getParent()->getReturnType();

  Value *RetValue;
  if (const IntegerType *ITy = dyn_cast<IntegerType>(RetType)) {

    if (ITy->getBitWidth() < 32) {
      RetValue = new TruncInst(Int32RetVal, RetType, "Trunc", FailBlock);

    } else if (ITy->getBitWidth() > 32) {
      RetValue = new SExtInst(Int32RetVal, RetType, "SExt", FailBlock);

    } else {
      RetValue = Int32RetVal;
    }

  } else if (RetType == Type::FloatTy || RetType == Type::DoubleTy) {
    RetValue = new SIToFPInst(Int32RetVal, RetType, "IntToFP", FailBlock);

  } else if (isa<PointerType>(RetType)) {
    RetValue = new IntToPtrInst(Int32RetVal, RetType, "IntToPtr", FailBlock);

  } else {
    assert(RetType == Type::VoidTy && "Unsupported return type");
    RetValue = 0;
  }
  ReturnInst::Create(RetValue, FailBlock);

  // Linking PrevBlock to TestBlock and OrigBlock
  BranchInst::Create(TestBlock, NextBlock, IsFailable, PrevBlock);

  // Linking TestBlock to FailBlock and OrigBlock
  BranchInst::Create(FailBlock, NextBlock, IsFailed, TestBlock);

  return true;
}


