
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Instructions.h"
#include "llvm/Constants.h"
#include "llvm/DerivedTypes.h"
#include "llvm/Function.h"
#include "llvm/ADT/DenseMap.h"
#include "llvm/Support/Compiler.h"
#include "llvm/Support/Streams.h"
#include "llvm/Support/Debug.h"
#include "llvm/Transforms/Instrumentation.h"
#include "llvm/Transforms/Scalar.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Transforms/Utils/Cloning.h"

#include "FaultInjectionUtils.h"

#include <sstream>
#include <map>
#include <vector>

using namespace llvm;
using namespace vertical;

// Helper functions
// ----------------

static Constant * GetDebugMsgPtr(Module *M, std::string Msg) {

  Constant *DebugMsg = ConstantArray::get("DEBUG: " + Msg + "\n");
  GlobalVariable *DebugMsgGbl = new GlobalVariable(
      DebugMsg->getType(), false, GlobalValue::InternalLinkage,
      DebugMsg, "Dbg", M);

  std::vector<Constant*> Indices(2);
  Indices[0] = Constant::getNullValue(Type::Int32Ty);
  Indices[1] = Constant::getNullValue(Type::Int32Ty);
  return ConstantExpr::getGetElementPtr(
      DebugMsgGbl, &Indices[0], Indices.size());
}


static Constant * GetPrintf(Module *M) {

  std::vector<const Type *> Params;
  Params.push_back(PointerType::getUnqual(Type::Int8Ty));
  FunctionType *PrintfTy = FunctionType::get(Type::VoidTy, Params, true);
  return M->getOrInsertFunction("printf", PrintfTy);
}


// Public functions
// ----------------


bool vertical::IsCallingExternOrIntrinsic(Instruction *I) {

  if (!(isa<InvokeInst>(I) || isa<CallInst>(I)))
    return false;

  CallSite CS = CallSite::get(I);

  // optimistically assume that function pointers are not calling externally
  // defined functions
  if (CS.getCalledFunction() && 
      (CS.getCalledFunction()->isIntrinsic() ||
       CS.getCalledFunction()->isDeclaration()))
    return true;
  else
    return false;
}


ReturnInst * vertical::CreateNullReturn(Instruction *I) {

  const Type * RetType = I->getParent()->getParent()->getReturnType();

  Value *RetValue = 0;
  if (RetType != Type::VoidTy)
    RetValue = Constant::getNullValue(RetType);

  return ReturnInst::Create(RetValue, I);
}


ReturnInst * vertical::CreateNullReturn(BasicBlock *BB) {

  const Type * RetType = BB->getParent()->getReturnType();

  Value *RetValue = 0;
  if (RetType != Type::VoidTy)
    RetValue = Constant::getNullValue(RetType);

  return ReturnInst::Create(RetValue, BB);
}


void vertical::CreateDebugMsg(const std::string &Msg, BasicBlock *BB) {

  Module *M = BB->getParent()->getParent();

  std::string FullMsg =
    BB->getParent()->getName() + "::" +
    BB->getName() + ": " + Msg;

  Constant *DebugMsgPtr = GetDebugMsgPtr(M, FullMsg);
  Constant *Printf = GetPrintf(M);

  CallInst::Create(Printf, DebugMsgPtr, "", BB);
}


void vertical::CreateDebugMsg(const std::string &Msg, Instruction *I) {

  Module *M = I->getParent()->getParent()->getParent();

  std::string FullMsg =
    I->getParent()->getParent()->getName() + "::" +
    I->getParent()->getName() + ": " + Msg;

  Constant *DebugMsgPtr = GetDebugMsgPtr(M, FullMsg);
  Constant *Printf = GetPrintf(M);

  CallInst::Create(Printf, DebugMsgPtr, "", I);
}


void vertical::ForkInsts(Module &M,
                         std::vector<Instruction *> &InstList,
                         std::map<Instruction *, Instruction *> &InstMap) {

  const Type *ATy;

  // Initialize the global probabilities array
  ATy = ArrayType::get(Type::FloatTy, InstList.size());
  GlobalVariable *ProbArray = new GlobalVariable(
      ATy, false, GlobalValue::InternalLinkage, Constant::getNullValue(ATy),
      "__POKE__ProbArray", &M);

  // Initialize the global countdown array
  ATy = ArrayType::get(Type::Int32Ty, InstList.size());
  GlobalVariable *CountdownArray = new GlobalVariable(
      ATy, false, GlobalValue::InternalLinkage, Constant::getNullValue(ATy),
      "__POKE__CountdownArray", &M);

  // Insert the initialization functions
  const PointerType *FloatPtr = PointerType::getUnqual(Type::FloatTy);
  const PointerType *Int32Ptr = PointerType::getUnqual(Type::Int32Ty);
  Constant *PokeInitFn = M.getOrInsertFunction("__POKE__Init", Type::VoidTy,
                                               (Type *)0);
  Constant *DiscardInitFn = M.getOrInsertFunction("__POKE__ForkedInit",
                                                  Type::VoidTy, FloatPtr,
                                                  Int32Ptr, Type::Int32Ty,
                                                  (Type *)0);

  Function *MainFn = M.getFunction("main");
  BasicBlock::iterator I = MainFn->begin()->begin();

  // Skip over any allocas in main's entry block.
  while (isa<AllocaInst>(I)) ++I;

  // call the poke initialization call
  CallInst::Create(PokeInitFn, "", I);

  // set the arguments for the discard initialization call and call it
  std::vector<Value*> Args(3);
  std::vector<Constant*> GEPIndices(2, Constant::getNullValue(Type::Int32Ty));
  Args[0] = ConstantExpr::getGetElementPtr(ProbArray, &GEPIndices[0],
                                           GEPIndices.size());
  Args[1] = ConstantExpr::getGetElementPtr(CountdownArray, &GEPIndices[0],
                                           GEPIndices.size());
  Args[2] = ConstantInt::get(Type::Int32Ty, InstList.size());
  CallInst::Create(DiscardInitFn, Args.begin(), Args.end(), "", I);


  for (unsigned i = 0; i < InstList.size(); i++) {

    Instruction *I = InstList[i];

    if (!I)
      continue;

    std::stringstream iString;
    iString << "id" << i;

    switch (I->getOpcode()) {
#define HANDLE_INST(NUM, OPCODE, CLASS) \
      case Instruction::OPCODE: \
          DOUT << "ForkInsts instrumenting "<< #OPCODE << " (" << i << \
              ") in " << I->getParent()->getName() << "\n"; \
      break;
#include "llvm/Instruction.def"
    }

    /*
     * The parent basic block of I is restructured as follows:
     *
     *         PrevBlock
     *          /    \
     *      TestBlock \
     *        /    \   \
     *    FailBlock OrigBlock
     *         \     /
     *       (NextBlock)
     *      
     *  PrevBlock contains all the logic leading up to I
     *  TestBlock contains code to check whether we should bother to generate
     *    a random number (i.e. checks whether the indexed value is non-zero)
     *  FailBlock contains the semantics for failure
     *  OrigBlock contains the original semantics
     *  NextBlock contains all the logic after I in the original basic block.
     *    (If I was a terminating instruction (e.g. a branch), NextBlock is
     *    not necessary)
     */

    BasicBlock *PrevBlock = I->getParent(); 
    BasicBlock *TestBlock; 
    BasicBlock *OrigBlock;
    BasicBlock *FailBlock;
    BasicBlock *NextBlock = PrevBlock->splitBasicBlock(
        I, iString.str() + ".next");

    Instruction *OrigI = I;
    Instruction *FailI;

    // Extending PrevBlock:
    // Here we perform the check to see if this instruction is injected with
    // failure.  This will be true if the countdown counter != 0.
    PrevBlock->getTerminator()->eraseFromParent();

    std::vector<Constant*> Indices(2);
    Indices[0] = Constant::getNullValue(Type::Int32Ty);
    Indices[1] = ConstantInt::get(Type::Int32Ty, i);
    Constant *CountdownPtr = 
      ConstantExpr::getGetElementPtr(CountdownArray, &Indices[0], Indices.size());
    Value *Count = new LoadInst(CountdownPtr, "Count", PrevBlock);

    ICmpInst *IsFailable = new ICmpInst(
        ICmpInst::ICMP_UGT, Count, Constant::getNullValue(Type::Int32Ty), 
        "Compare" + iString.str(), PrevBlock);

    // Creating TestBlock:
    // In TestBlock we decrement the counter and see if it is equal to 0. If
    // so, we take the failure path.
    TestBlock = BasicBlock::Create(iString.str() + ".test",
                                   NextBlock->getParent(),
                                   NextBlock);

    Value *DecCount =
      BinaryOperator::CreateSub(Count, ConstantInt::get(Type::Int32Ty, 1),
                                "DecCounter", TestBlock);

    new StoreInst(DecCount, CountdownPtr, TestBlock);

    ICmpInst *IsFail = new ICmpInst(ICmpInst::ICMP_EQ, DecCount,
                                    Constant::getNullValue(Type::Int32Ty),
                                    "IsFail", TestBlock);


    // Creating OrigBlock:
    // This is our unchanged execution path. 
    if (isa<TerminatorInst>(I)) {
      OrigBlock = NextBlock;

    } else {
      OrigBlock = BasicBlock::Create(iString.str(), NextBlock->getParent(),
                                     NextBlock);
      Instruction *OrigBranch = BranchInst::Create(NextBlock, OrigBlock);

      I->moveBefore(OrigBranch);  
    }

    // Creating FailBlock:
    // This is the failure path.  We simply duplicate the behavior for now.
    // We also generate a new countdown value.
    DenseMap<const Value *, Value *> ValueMap;
    FailBlock = CloneBasicBlock(OrigBlock, ValueMap, ".failed",
                               OrigBlock->getParent());

    // this is the failed version of the instruction
    FailI = FailBlock->getFirstNonPHI();

    // get handle to random number generator
    Constant *RandIntExp = M.getOrInsertFunction("__POKE__RandUIntExp",
                                                 Type::Int32Ty, Type::FloatTy,
                                                 (Type *)0);

    // get the failure probability
    Constant *ProbPtr = 
      ConstantExpr::getGetElementPtr(ProbArray, &Indices[0], Indices.size());
    Value *Prob = new LoadInst(ProbPtr, "Prob", FailI);

    // call the random number generator to get the new countdown value
    CallInst *NewCount = CallInst::Create(RandIntExp, Prob, "Rand",
                                                FailI);

    // Update the counter 
    new StoreInst(NewCount, CountdownPtr, FailI);

    // if we are considering a terminating inst (e.g. branch), we need to
    // update successors to accept both FailBlock and OrigBlock as potential
    // source points
    if (isa<TerminatorInst>(I)) {

      for (succ_iterator S = succ_begin(FailBlock); S != succ_end(FailBlock);
           ++S) {

        // Loop over any phi nodes in the successor basic blocks, updating
        // the BB field of incoming values...
        BasicBlock *Successor = *S;
        PHINode *PN;

        for (BasicBlock::iterator I = Successor->begin();
             (PN = dyn_cast<PHINode>(I)); ++I) {

          int IDX = PN->getBasicBlockIndex(OrigBlock);
          if (IDX != -1) {
            // the incoming value will be the same for FailBlock as it was for
            // OrigBlock
            PN->addIncoming(PN->getIncomingValueForBlock(OrigBlock),
                            FailBlock);
          }
        }
      }
    }

    // Linking PrevBlock to TestBlock and OrigBlock
    BranchInst::Create(TestBlock, OrigBlock, IsFailable, PrevBlock);

    // Linking TestBlock to FailBlock and OrigBlock
    BranchInst::Create(FailBlock, OrigBlock, IsFail, TestBlock);

    // if the value is used, create PHI node in previously split off NextBlock
    if (!I->use_empty()) {

      // insert a PHINode before the successor block
      PHINode *PN = PHINode::Create(
          I->getType(), "IncomingValue",
          NextBlock->begin());

      // Replacing uses of I with PHI value
      I->replaceAllUsesWith(PN);

      PN->addIncoming(OrigI, OrigBlock);
      PN->addIncoming(FailI, FailBlock);
    }

    // insert bad instruction into the failed instruction list
    InstMap[OrigI] = FailI;
  }
}

