
#include "llvm/GlobalVariable.h"
#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Support/Debug.h"

#include "FaultInjectionUtils.h"
#include "InstructionInjection.h"
#include "LoopUpdateVars.h"
#include "Snapshot.h"

#include <set>

using namespace llvm;
using namespace vertical;


namespace {
  class FunctionDiscardInjection : public InstructionInjection {
  private:
    bool runOnFunction(Function &F);
    bool doInitialization(Module &M);
    bool doFinalization(Module &M);
    void getAnalysisUsage(AnalysisUsage &AU) const;
    bool isInstrumented(Instruction *I);

  public:
    static char ID;
    FunctionDiscardInjection() : InstructionInjection((intptr_t)&ID) {}

    GlobalVariable *RCArray;
  };
}


char FunctionDiscardInjection::ID = 0;
static RegisterPass<FunctionDiscardInjection> X(
    "func-discard", "Function-Discard Fault Injection");


void FunctionDiscardInjection::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequired<Snapshot>();
  InstructionInjection::getAnalysisUsage(AU);
}


bool FunctionDiscardInjection::isInstrumented(Instruction *I) {
  switch (I->getOpcode()) {

    // call instructions (ADDED!)
    case Instruction::Call:
    // ptr computation/access
    case Instruction::GetElementPtr:
    case Instruction::Load:
    case Instruction::Store:
    // comparison
    case Instruction::ICmp:
    case Instruction::FCmp:
    // control
    case Instruction::Select:
    // arithmetic/logic
    case Instruction::Add:
    case Instruction::Sub:
    case Instruction::Mul:
    case Instruction::UDiv:
    case Instruction::SDiv:
    case Instruction::FDiv:
    case Instruction::URem:
    case Instruction::SRem:
    case Instruction::FRem:
    case Instruction::Shl:
    case Instruction::LShr:
    case Instruction::AShr:
    case Instruction::And:
    case Instruction::Or:
    case Instruction::Xor:
    // casts
    case Instruction::FPToUI:
    case Instruction::FPToSI:
    case Instruction::UIToFP:
    case Instruction::SIToFP:
    // casts (ADDED!)
    case Instruction::Trunc:
    case Instruction::ZExt:
    case Instruction::SExt:
    case Instruction::FPTrunc:
    case Instruction::FPExt:
    case Instruction::PtrToInt:
    case Instruction::IntToPtr:
    case Instruction::BitCast:
      // supported
      return true;

    default:
      // unsupported
      return false;
  }
}


bool FunctionDiscardInjection::doInitialization(Module &M) {

  const Type *IntATy = ArrayType::get(Type::Int32Ty, M.size());
  RCArray = new GlobalVariable(
      IntATy, false, GlobalValue::InternalLinkage,
      Constant::getNullValue(IntATy),
      "__POKE__RCArray", &M);

  return InstructionInjection::doInitialization(M);
}


bool FunctionDiscardInjection::doFinalization(Module &M) {

  // Insert the initialization functions
  const PointerType *UIntPtr  = PointerType::getUnqual(Type::Int32Ty);
  Constant *DiscardInitFn = M.getOrInsertFunction("__POKE__FuncDiscardInit",
                                                  Type::VoidTy, UIntPtr,
                                                  Type::Int32Ty, (Type *)0);

  Function *MainFn = M.getFunction("main");
  BasicBlock::iterator I = MainFn->begin()->begin();

  // Skip over any allocas in main's entry block.
  while (isa<AllocaInst>(I)) ++I;

  // set the arguments for the discard initialization call and call it
  std::vector<Value*> Args(2);
  std::vector<Constant*> GEPIndices(2, Constant::getNullValue(Type::Int32Ty));
  Args[0] = ConstantExpr::getGetElementPtr(RCArray, &GEPIndices[0],
                                           GEPIndices.size());
  Args[1] = ConstantInt::get(Type::Int32Ty, M.size());
  CallInst::Create(DiscardInitFn, Args.begin(), Args.end(), "", I);

  InstructionInjection::doFinalization(M);

  return true;
}



bool FunctionDiscardInjection::runOnFunction(Function &F) {

  Snapshot &SS = getAnalysis<Snapshot>();

  forkInsts(F);

  for (std::map<Instruction *, Instruction *>::iterator IMI =
       InjectionMap.begin(); IMI != InjectionMap.end(); ++IMI) {

    Instruction *OrigI = IMI->first;
    Instruction *FailI = IMI->second;

    DOUT << "FunctionDiscardInjection: " << SS.IIMap[OrigI]->StringId << "\n";

    BasicBlock *FailBlock = FailI->getParent();

    // remove successor PHI entries
    typedef GraphTraits<BasicBlock *> BlockTraits;
    for (BlockTraits::ChildIteratorType SI = BlockTraits::child_begin(FailBlock);
         SI != BlockTraits::child_end(FailBlock); ++SI) {

      BasicBlock *Next = *SI;
      PHINode *PN;
      for (BasicBlock::iterator I = Next->begin();
           (PN = dyn_cast<PHINode>(I)); ++I) {

        int Idx = PN->getBasicBlockIndex(FailBlock);
        while (Idx != -1) {
          PN->removeIncomingValue(Idx);
          Idx = PN->getBasicBlockIndex(FailBlock);
        }
      }
    }

    // replace the terminator with a return instruction
    FailBlock->getTerminator()->eraseFromParent();
  
    // insert the return instructions
    std::vector<Constant*> RCIndices(2);
    RCIndices[0] = Constant::getNullValue(Type::Int32Ty);
    RCIndices[1] = ConstantInt::get(Type::Int32Ty, SS.FI->Id);
    Constant *RCElementPtr = 
      ConstantExpr::getGetElementPtr(RCArray, &RCIndices[0], RCIndices.size());
    Value *Int32RetVal = new LoadInst(RCElementPtr, "RetVal", FailBlock);

    const Type *RetType = F.getReturnType();

    Value *RetValue;
    if (const IntegerType *ITy = dyn_cast<IntegerType>(RetType)) {

      if (ITy->getBitWidth() < 32) {
        RetValue = new TruncInst(Int32RetVal, RetType, "Trunc", FailBlock);

      } else if (ITy->getBitWidth() > 32) {
        RetValue = new SExtInst(Int32RetVal, RetType, "SExt", FailBlock);

      } else {
        RetValue = Int32RetVal;
      }

    } else if (RetType == Type::FloatTy || RetType == Type::DoubleTy) {
      RetValue = new SIToFPInst(Int32RetVal, RetType, "IntToFP", FailBlock);

    } else if (isa<PointerType>(RetType)) {
      RetValue = new IntToPtrInst(Int32RetVal, RetType, "IntToPtr", FailBlock);

    } else {
      assert(RetType == Type::VoidTy && "Unsupported return type");
      RetValue = 0;
    }

    ReturnInst::Create(RetValue, FailBlock);
  }

  return true;
}



