
#include "llvm/Constants.h"
#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Analysis/Dominators.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/Streams.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include "llvm/Transforms/Utils/Cloning.h"

#include <sstream>
#include <map>

#include "InstructionInjection.h"
#include "Snapshot.h"

using namespace llvm;
using namespace vertical;

static cl::opt<bool> ProfilingHooks(
    "profiling-hooks", cl::Hidden, 
    cl::desc("program calls profiling hooks"));


char InstructionInjection::ID = 0;

void InstructionInjection::getAnalysisUsage(AnalysisUsage &AU) const {
  // AU.addPreserved<DominanceFrontier>();
  // AU.addPreserved<DominatorTree>();
  // AU.addPreserved<LoopInfo>();
  AU.addRequired<Snapshot>();
}


bool InstructionInjection::doInitialization(Module &M) {

  InstCount = 0;
  for (Module::iterator F = M.begin(); F != M.end(); ++F)
    for (Function::iterator BB = F->begin(); BB != F->end(); ++BB)
      InstCount += BB->size();

  // Initialize the global probabilities array
  const Type *ATy;
  ATy = ArrayType::get(Type::FloatTy, InstCount);
  ProbArray = new GlobalVariable(
      ATy, false, GlobalValue::InternalLinkage, Constant::getNullValue(ATy),
      "__POKE__ProbArray", &M);

  // Initialize the global countdown array
  ATy = ArrayType::get(Type::Int32Ty, InstCount);
  CountdownArray = new GlobalVariable(
      ATy, false, GlobalValue::InternalLinkage, Constant::getNullValue(ATy),
      "__POKE__CountdownArray", &M);

  return false;
}


bool InstructionInjection::doFinalization(Module &M) {

  // Insert the initialization functions
  const PointerType *FloatPtr = PointerType::getUnqual(Type::FloatTy);
  const PointerType *Int32Ptr = PointerType::getUnqual(Type::Int32Ty);

  Function *MainFn = M.getFunction("main");
  BasicBlock::iterator I = MainFn->begin()->begin();

  // Skip over any allocas in main's entry block.
  while (isa<AllocaInst>(I)) ++I;

  // call the poke initialization call
  Constant *PokeInitFn = M.getOrInsertFunction(
      "__POKE__Init", Type::VoidTy, (Type *)0);
  CallInst::Create(PokeInitFn, "", I);

  // set the arguments for the discard initialization call and call it
  std::vector<Value*> Args(3);
  std::vector<Constant*> GEPIndices(2, Constant::getNullValue(Type::Int32Ty));
  Args[0] = ConstantExpr::getGetElementPtr(ProbArray, &GEPIndices[0],
                                           GEPIndices.size());
  Args[1] = ConstantExpr::getGetElementPtr(CountdownArray, &GEPIndices[0],
                                           GEPIndices.size());
  Args[2] = ConstantInt::get(Type::Int32Ty, InstCount);

  Constant *DiscardInitFn = M.getOrInsertFunction(
      "__POKE__ForkedInit", Type::VoidTy, FloatPtr,
      Int32Ptr, Type::Int32Ty, (Type *)0);
  CallInst::Create(DiscardInitFn, Args.begin(), Args.end(), "", I);

  // add InstCounter call
  Constant *CounterInitFn = M.getOrInsertFunction(
      "__POKE__InstCounterInit", Type::VoidTy, (Type *)0);
  CallInst::Create(CounterInitFn, "", I);

  if (!ProfilingHooks) {
    Constant *ProfInitFn = M.getOrInsertFunction("__POKE__ProfilingBegin",
                                             Type::VoidTy, (Type *)0);
    CallInst::Create(ProfInitFn, "", I);
  }

  return true;
}

void InstructionInjection::addBlockToAnalyses(BasicBlock *BB,
                                              BasicBlock *Dom)
{
  LoopInfo      *LIN = getAnalysisIfAvailable<LoopInfo>();
  DominatorTree *DT  = getAnalysisIfAvailable<DominatorTree>();

  if (LIN) {
    if (Loop *L = LIN->getLoopFor(Dom)) {
      L->addBasicBlockToLoop(BB, LIN->getBase());
    }
  }

  if (DT)
    DT->addNewBlock(BB, Dom);
}

void InstructionInjection::forkInsts(Function &F) {

  InjectionMap.clear();

  Snapshot &SS  = getAnalysis<Snapshot>();

  DOUT << "forkInsts: " << F.getName() << "\n";

  for (Snapshot::ii_iterator II = SS.ii_begin(); II != SS.ii_end(); ++II) {
    Instruction *I = (*II)->I;

    if (!isInstrumented(I))
      continue;

    // skip exception handler calls
    if (CallInst *Call = dyn_cast<CallInst>(I)) {
      if (Function *F = Call->getCalledFunction()) {
        if (F->getName().substr(0, 7).compare("llvm.eh") == 0) {
          DOUT << "  Skipping call to " << F->getName() << "\n";
          continue;
        }
      }
    }

    std::stringstream idString;
    idString << "id" << (*II)->Id;

#if 0 
    switch (I->getOpcode()) {
#define HANDLE_INST(NUM, OPCODE, CLASS) \
      case Instruction::OPCODE: \
          DOUT << "  Forking "<< #OPCODE << " (" << (*II)->Id << \
              "):" << (*II)->StringId << "\n"; \
      break;
#include "llvm/Instruction.def"
    }
#endif

    /*
     * The parent basic block of I is restructured as follows:
     *
     *         PrevBlock
     *          /    \
     *      TestBlock \
     *        /    \   \
     *    FailBlock OrigBlock
     *         \     /
     *       (NextBlock)
     *      
     *  PrevBlock contains all the logic leading up to I
     *  TestBlock contains code to check whether we should bother to generate
     *    a random number (i.e. checks whether the indexed value is non-zero)
     *  FailBlock contains the semantics for failure
     *  OrigBlock contains the original semantics
     *  NextBlock contains all the logic after I in the original basic block.
     *    (If I was a terminating instruction (e.g. a branch), NextBlock is
     *    not necessary) -- it will contain the code to check the next
     *    instruction for failure
     */

    BasicBlock *PrevBlock = I->getParent(); 
    BasicBlock *TestBlock; 
    BasicBlock *OrigBlock;
    BasicBlock *FailBlock;
    BasicBlock *NextBlock = SplitBlock(PrevBlock, I, this);

    Instruction *OrigI = I;
    Instruction *FailI;

    // Extending PrevBlock:
    // Here we perform the check to see if this instruction is injected with
    // failure.  This will be true if the countdown counter != 0.
    PrevBlock->getTerminator()->eraseFromParent();

    std::vector<Constant*> Indices(2);
    Indices[0] = Constant::getNullValue(Type::Int32Ty);
    Indices[1] = ConstantInt::get(Type::Int32Ty, (*II)->Id);
    Constant *CountdownPtr = 
      ConstantExpr::getGetElementPtr(CountdownArray, &Indices[0], Indices.size());
    Value *Count = new LoadInst(CountdownPtr, "Count", PrevBlock);

    ICmpInst *IsFailable = new ICmpInst(
        ICmpInst::ICMP_UGT, Count, Constant::getNullValue(Type::Int32Ty), 
        "Compare" + idString.str(), PrevBlock);

    // Creating TestBlock:
    // In TestBlock we decrement the counter and see if it is equal to 0. If
    // so, we take the failure path.
    TestBlock = BasicBlock::Create(idString.str() + ".Test",
                                   NextBlock->getParent(),
                                   NextBlock);
    addBlockToAnalyses(TestBlock, PrevBlock);

    Value *DecCount =
      BinaryOperator::CreateSub(Count, ConstantInt::get(Type::Int32Ty, 1),
                                "DecCounter", TestBlock);

    new StoreInst(DecCount, CountdownPtr, TestBlock);

    ICmpInst *IsFail = new ICmpInst(ICmpInst::ICMP_EQ, DecCount,
                                    Constant::getNullValue(Type::Int32Ty),
                                    "IsFail", TestBlock);


    // Creating OrigBlock:
    // This is our unchanged execution path. 
    if (isa<TerminatorInst>(I)) {
      OrigBlock = NextBlock;

    } else {
      OrigBlock = BasicBlock::Create(idString.str(), NextBlock->getParent(),
                                     NextBlock);
      addBlockToAnalyses(OrigBlock, PrevBlock);

      Instruction *OrigBranch = BranchInst::Create(NextBlock, OrigBlock);

      I->moveBefore(OrigBranch);  
    }

    // Creating FailBlock:
    // This is the failure path.  We simply duplicate the behavior for now.
    // We also generate a new countdown value.
    DenseMap<const Value *, Value *> ValueMap;
    FailBlock = CloneBasicBlock(OrigBlock, ValueMap, ".Failed",
                               OrigBlock->getParent());

    addBlockToAnalyses(FailBlock, PrevBlock);

    // this is the failed version of the instruction
    FailI = FailBlock->getFirstNonPHI();

    // get handle to random number generator
    Module *M = F.getParent();
    Constant *RandIntExp = M->getOrInsertFunction("__POKE__RandUIntExp",
                                                  Type::Int32Ty, Type::FloatTy,
                                                  (Type *)0);

    // get the failure probability
    Constant *ProbPtr = 
      ConstantExpr::getGetElementPtr(ProbArray, &Indices[0], Indices.size());
    Value *Prob = new LoadInst(ProbPtr, "Prob", FailI);

    // call the random number generator to get the new countdown value
    CallInst *NewCount = CallInst::Create(RandIntExp, Prob, "Rand",
                                                FailI);

    // Update the counter 
    new StoreInst(NewCount, CountdownPtr, FailI);

    // if we are considering a terminating inst (e.g. branch), we need to
    // update successors to accept both FailBlock and OrigBlock as potential
    // source points
    if (isa<TerminatorInst>(I)) {

      for (succ_iterator S = succ_begin(FailBlock); S != succ_end(FailBlock);
           ++S) {

        // Loop over any phi nodes in the successor basic blocks, updating
        // the BB field of incoming values...
        BasicBlock *Successor = *S;
        PHINode *PN;

        for (BasicBlock::iterator I = Successor->begin();
             (PN = dyn_cast<PHINode>(I)); ++I) {

          int IDX = PN->getBasicBlockIndex(OrigBlock);
          if (IDX != -1) {
            // the incoming value will be the same for FailBlock as it was for
            // OrigBlock
            PN->addIncoming(PN->getIncomingValueForBlock(OrigBlock),
                            FailBlock);
          }
        }
      }
    }

    // Linking PrevBlock to TestBlock and OrigBlock
    BranchInst::Create(TestBlock, OrigBlock, IsFailable, PrevBlock);

    // Linking TestBlock to FailBlock and OrigBlock
    BranchInst::Create(FailBlock, OrigBlock, IsFail, TestBlock);

    // if the value is used, create PHI node in previously split off NextBlock
    if (!I->use_empty()) {

      // insert a PHINode before the successor block
      PHINode *PN = PHINode::Create(
          I->getType(), "IncomingValue",
          NextBlock->begin());

      // Replacing uses of I with PHI value
      I->replaceAllUsesWith(PN);

      PN->addIncoming(OrigI, OrigBlock);
      PN->addIncoming(FailI, FailBlock);
    }

    // insert forked version of the instruction into the injection map
    InjectionMap[OrigI] = FailI;
  }
}

void InstructionInjection::forkInsts(Function &F,
    std::vector<Instruction *>                &InjectList,
    std::map<Instruction *, Instruction *>    &InjectMap) {

  Snapshot &SS  = getAnalysis<Snapshot>();

  DOUT << "forkInsts: " << F.getName() << "\n";

  for (std::vector<Instruction *>::iterator II = InjectList.begin(),
       IE = InjectList.end(); II != IE; ++II) {

    Instruction *I = *II;
    Module *M = I->getParent()->getParent()->getParent();

    assert(!isa<PHINode>(I) && "cannot inject into PHINode");
    assert(!isa<InvokeInst>(I) && "cannot inject into InvokeInst (yet)");

    InstructionInfo *IInf = SS.IIMap[I];

    std::stringstream idString;
    idString << "id" << IInf->Id;

    switch (I->getOpcode()) {
#define HANDLE_INST(NUM, OPCODE, CLASS) \
      case Instruction::OPCODE: \
          DOUT << "  Forking "<< #OPCODE << " (" << IInf->Id << \
              "):" << IInf->StringId << "\n"; \
      break;
#include "llvm/Instruction.def"
    }

    /*
     * The parent basic block of I is restructured as follows:
     *
     *         PrevBlock
     *          /    \
     *      TestBlock \
     *        /    \   \
     *    FailBlock OrigBlock
     *         \     /
     *       (NextBlock)
     *      
     *  PrevBlock contains all the logic leading up to I
     *  TestBlock contains code to check whether we should bother to generate
     *    a random number (i.e. checks whether the indexed value is non-zero)
     *  FailBlock contains the semantics for failure
     *  OrigBlock contains the original semantics
     *  NextBlock contains all the logic after I in the original basic block.
     *    (If I was a terminating instruction (e.g. a branch), NextBlock is
     *    not necessary) -- it will contain the code to check the next
     *    instruction for failure
     */

    BasicBlock *PrevBlock = I->getParent(); 
    BasicBlock *TestBlock; 
    BasicBlock *OrigBlock;
    BasicBlock *FailBlock;
    BasicBlock *NextBlock = SplitBlock(PrevBlock, I, this);

    Instruction *OrigI = I;
    Instruction *FailI;

    // Extending PrevBlock:
    // Here we perform the check to see if this instruction is injected with
    // failure.  This will be true if the countdown counter != 0.
    PrevBlock->getTerminator()->eraseFromParent();

    std::vector<Constant*> Indices(2);
    Indices[0] = Constant::getNullValue(Type::Int32Ty);
    Indices[1] = ConstantInt::get(Type::Int32Ty, IInf->Id);
    Constant *CountdownPtr = 
      ConstantExpr::getGetElementPtr(CountdownArray, &Indices[0], Indices.size());
    Value *Count = new LoadInst(CountdownPtr, "Count", PrevBlock);

    ICmpInst *IsFailable = new ICmpInst(
        ICmpInst::ICMP_UGT, Count, Constant::getNullValue(Type::Int32Ty), 
        "Compare" + idString.str(), PrevBlock);

    // Creating TestBlock:
    // In TestBlock we decrement the counter and see if it is equal to 0. If
    // so, we take the failure path.
    TestBlock = BasicBlock::Create(idString.str() + ".Test",
                                   NextBlock->getParent(),
                                   NextBlock);
    addBlockToAnalyses(TestBlock, PrevBlock);

    Value *DecCount =
      BinaryOperator::CreateSub(Count, ConstantInt::get(Type::Int32Ty, 1),
                                "DecCounter", TestBlock);

    new StoreInst(DecCount, CountdownPtr, TestBlock);

    ICmpInst *IsFail = new ICmpInst(ICmpInst::ICMP_EQ, DecCount,
                                    Constant::getNullValue(Type::Int32Ty),
                                    "IsFail", TestBlock);


    // Creating OrigBlock:
    // This is our unchanged execution path. 
    if (isa<TerminatorInst>(I)) {
      OrigBlock = NextBlock;

    } else {
      OrigBlock = BasicBlock::Create(idString.str(), NextBlock->getParent(),
                                    NextBlock);
      addBlockToAnalyses(OrigBlock, PrevBlock);

      Instruction *OrigBranch = BranchInst::Create(NextBlock, OrigBlock);

      I->moveBefore(OrigBranch);  
    }

    // Creating FailBlock:
    // This is the failure path.  We simply duplicate the behavior for now.
    // We also generate a new countdown value.
    DenseMap<const Value *, Value *> ValueMap;
    FailBlock = CloneBasicBlock(OrigBlock, ValueMap, ".Failed",
                               OrigBlock->getParent());
    addBlockToAnalyses(FailBlock, PrevBlock);

    // this is the failed version of the instruction
    FailI = FailBlock->getFirstNonPHI();

    // get handle to random number generator
    Constant *RandIntExp = M->getOrInsertFunction("__POKE__RandUIntExp",
                                                  Type::Int32Ty, Type::FloatTy,
                                                  (Type *)0);

    // get the failure probability
    Constant *ProbPtr = 
      ConstantExpr::getGetElementPtr(ProbArray, &Indices[0], Indices.size());
    Value *Prob = new LoadInst(ProbPtr, "Prob", FailI);

    // call the random number generator to get the new countdown value
    CallInst *NewCount = CallInst::Create(RandIntExp, Prob, "Rand",
                                                FailI);

    // Update the counter 
    new StoreInst(NewCount, CountdownPtr, FailI);

    // Linking PrevBlock to TestBlock and OrigBlock
    BranchInst::Create(TestBlock, OrigBlock, IsFailable, PrevBlock);

    // Linking TestBlock to FailBlock and OrigBlock
    BranchInst::Create(FailBlock, OrigBlock, IsFail, TestBlock);

    // if the value is used, create PHI node in previously split off NextBlock
    if (!I->use_empty()) {

      // insert a PHINode before the successor block
      PHINode *PN = PHINode::Create(
          I->getType(), "IncomingValue",
          NextBlock->begin());

      // Replacing uses of I with PHI value
      I->replaceAllUsesWith(PN);

      PN->addIncoming(OrigI, OrigBlock);
      PN->addIncoming(FailI, FailBlock);
    }

    // insert forked version of the instruction into the injection map
    InjectMap[OrigI] = FailI;
  }

  DOUT << "addProfiling: " << F.getName() << "\n";

  // allocate the instruction counter (will be register-allocated by mem2reg)
  BasicBlock *Entry = F.begin();
  BasicBlock::iterator InsertPos = Entry->begin();
  while (isa<AllocaInst>(InsertPos)) ++InsertPos;

  Value *CounterPtr = new AllocaInst(Type::Int32Ty, "CounterPtr", InsertPos);
  new StoreInst(ConstantInt::getNullValue(Type::Int32Ty), CounterPtr,
                InsertPos);

  // now iterate over the blocks to populate the counter
  Module *M = F.getParent();
  Constant *InstCounterAccumFn = M->getOrInsertFunction(
      "__POKE__InstCounterAccum", Type::VoidTy, Type::Int32Ty, NULL);

  for (Snapshot::bi_iterator BII = SS.bi_begin(); BII != SS.bi_end(); ++BII) {
    BlockInfo *BI = *BII;
    BasicBlock *BB = BI->BB;
    DOUT << "  " << BI->StringId << "\n";

    unsigned counter = 0;
    for (BlockInfo::iterator III = BI->begin(), IIE = BI->end(); III != IIE;
         ++III) {
      InstructionInfo *II = *III;
      Instruction *I = II->I;

      if (!InjectMap.count(I)) {
        counter++;
      } else {
        Value *OldCounter = new LoadInst(CounterPtr, "OldCounter", I);
        Value *NewCounter = BinaryOperator::CreateAdd(
            OldCounter, ConstantInt::get(Type::Int32Ty, 1),
            "NewCounter", I);
        new StoreInst(NewCounter, CounterPtr, I);
      }
    }

    if (counter > 0) {
      Instruction *InsertPos = BB->getTerminator();
      Value *OldCounter = new LoadInst(CounterPtr, "OldCounter", InsertPos);
      Value *NewCounter = BinaryOperator::CreateAdd(
          OldCounter, ConstantInt::get(Type::Int32Ty, counter),
          "NewCounter", InsertPos);
      new StoreInst(NewCounter, CounterPtr, InsertPos);
    }
  }

  // find all return instructions and prepend the retire call
  for (Function::iterator BB = F.begin(); BB != F.end(); ++BB) {
    for (BasicBlock::iterator I = BB->begin(); I != BB->end(); ++I) {
      if (isa<ReturnInst>(I)) {
        Value *Counter = new LoadInst(CounterPtr, "Counter", I);
        CallInst::Create(InstCounterAccumFn, Counter, "", I);
      }
    }
  }
}

