
#include "llvm/GlobalVariable.h"
#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Support/Debug.h"

#include "Criticality.h"
#include "FaultInjectionUtils.h"
#include "InstructionInjection.h"
#include "LoopUpdateVars.h"
#include "Snapshot.h"

#include <set>

using namespace llvm;
using namespace vertical;


namespace {
  class LoopDiscardInjection : public InstructionInjection {
  private:
    bool runOnFunction(Function &F);
    void getAnalysisUsage(AnalysisUsage &AU) const;
    bool isInstrumented(Instruction *I);

  public:
    static char ID;
    LoopDiscardInjection() : InstructionInjection((intptr_t)&ID) {}

  private:
    std::set<Instruction *> AlreadyHandled;
  };
}


char LoopDiscardInjection::ID = 0;
static RegisterPass<LoopDiscardInjection> X(
    "loop-discard", "Loop-Discard Fault Injection");


void LoopDiscardInjection::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequired<LoopUpdateVars>();
  AU.addRequired<Criticality>();
  AU.addRequired<Snapshot>();
  InstructionInjection::getAnalysisUsage(AU);
}


bool LoopDiscardInjection::isInstrumented(Instruction *I) {
  switch (I->getOpcode()) {

    // memory creation (ADDED!)
    case Instruction::Malloc:
    // call instructions (ADDED!)
    case Instruction::Call:
    // ptr computation/access
    case Instruction::GetElementPtr:
    case Instruction::Load:
    case Instruction::Store:
    // comparison
    case Instruction::ICmp:
    case Instruction::FCmp:
    // control
    case Instruction::Select:
    // arithmetic/logic
    case Instruction::Add:
    case Instruction::Sub:
    case Instruction::Mul:
    case Instruction::UDiv:
    case Instruction::SDiv:
    case Instruction::FDiv:
    case Instruction::URem:
    case Instruction::SRem:
    case Instruction::FRem:
    case Instruction::Shl:
    case Instruction::LShr:
    case Instruction::AShr:
    case Instruction::And:
    case Instruction::Or:
    case Instruction::Xor:
    // casts
    case Instruction::FPToUI:
    case Instruction::FPToSI:
    case Instruction::UIToFP:
    case Instruction::SIToFP:
    // casts (ADDED!)
    case Instruction::Trunc:
    case Instruction::ZExt:
    case Instruction::SExt:
    case Instruction::FPTrunc:
    case Instruction::FPExt:
    case Instruction::PtrToInt:
    case Instruction::IntToPtr:
    case Instruction::BitCast:
      // supported
      return true;

    default:
      // unsupported
      return false;
  }
}


bool LoopDiscardInjection::runOnFunction(Function &F) {

  Criticality    &CRT = getAnalysis<Criticality>();
  LoopUpdateVars &LUV = getAnalysis<LoopUpdateVars>();
  Snapshot       &SS  = getAnalysis<Snapshot>();

  Constant *Off = ConstantInt::get(Type::Int1Ty, 0);
  Constant *On  = ConstantInt::get(Type::Int1Ty, 1);

  forkInsts(F);

  for (LoopUpdateVars::iterator UI = LUV.begin(), UE = LUV.end();
       UI != UE; ++UI) {

    UpdateInfo *U = *UI;
    BasicBlock *H  = U->Parent->getHeader();

    assert(SS.IIMap.count(U->Update) != 0);
    DOUT << "LoopDiscardInjection: " << SS.IIMap[U->Update]->StringId << "\n";
    assert(InjectionMap.count(U->Update) != 0);

    if (CRT.CriticalityMap[U->Update] != 0) {
      // this instruction is critical -- don't do it!
      DOUT << "  Critical! Skipping....\n";
      continue;
    }

    if (AlreadyHandled.count(U->Update)) {
      // already done in earlier, inner loop
      DOUT << "  Already handled! Skipping....\n";
      continue; 
    }

    // we'll claim it
    AlreadyHandled.insert(U->Update);

    // insert the bypass bit variable
    BasicBlock::iterator InsertPos = 
      U->Update->getParent()->getParent()->begin()->begin();
    while (isa<PHINode>(InsertPos) || isa<AllocaInst>(InsertPos)) ++InsertPos;
    AllocaInst *BypassBit = new AllocaInst(Type::Int1Ty, "BypassLV", InsertPos); 

    // reset it at the beginning of each loop iteration
    new StoreInst(Off, BypassBit, H->getFirstNonPHI());

    // iterate over the dependents of this function and mark them if they are
    // not critical
    for (UpdateInfo::iterator II = U->begin(), IE = U->end(); II != IE; ++II)
    {
      Instruction *OrigI = *II;

      if (InjectionMap.count(OrigI) == 0) {
        // this instruction is not instrumentable -- skip it
        DOUT << "  Skipping non-instrumentable instruction " <<
          SS.IIMap[OrigI]->StringId << "\n";

      } else if (CRT.CriticalityMap[OrigI] != 0) {
        // this instruction is critical -- skip it
        DOUT << "  Skipping critical instruction " <<
          SS.IIMap[OrigI]->StringId << "\n";

      } else {
        DOUT << "  Marking " << SS.IIMap[OrigI]->StringId << "\n";
        Instruction *FailI = InjectionMap[OrigI];
        new StoreInst(On, BypassBit, FailI);
      }
    }

    DOUT << "  Creating unconditional bypass for source" << "\n";
    // for the U->Update instruction, first implement the failure behavior
    // (unconditional bypass)
    Instruction *FailUpdate = InjectionMap[U->Update];
    if (isa<StoreInst>(U->Update)) {
      // simply erase the store instruction
      FailUpdate->eraseFromParent();
    } else {
      // otherwise bypass source PHINode
      FailUpdate->replaceAllUsesWith(U->Source);
    }

    DOUT << "  Creating conditional bypass for source" << "\n";
    // now implement the non-failure bypass (conditional bypass) --
    // we will implement the bypass using a PHI node
    BasicBlock *PreBlock  = U->Update->getParent();
    BasicBlock *PostBlock = PreBlock->splitBasicBlock(
        U->Update, PreBlock->getName() + ".PHI"); 

    // this is the block containing the original functionality which we will
    // stick in the middle
    BasicBlock *OrigBlock = BasicBlock::Create(
        PreBlock->getName() + ".Orig", PreBlock->getParent());
    BranchInst *OrigBranch = BranchInst::Create(PostBlock, OrigBlock);
    U->Update->moveBefore(OrigBranch);

    // turn the unconditional branch in PreBlock into a conditional branch
    PreBlock->getTerminator()->eraseFromParent();
    Value *IsBypass = new LoadInst(BypassBit, "BypassBit", PreBlock);
    BranchInst::Create(PostBlock, OrigBlock, IsBypass, PreBlock);

    if (!isa<StoreInst>(U->Update)) {
      // insert a PHINode before PostBlock
      PHINode *PN = PHINode::Create(U->Update->getType(), "Bypass",
                                    PostBlock->begin());

      // Replace uses of U->Update with PHI value
      U->Update->replaceAllUsesWith(PN);

      // Now update the PHINode (don't do this before replaceAllUsesWith or we
      // will update the PN use as well, which is not correct)
      PN->addIncoming(U->Source, PreBlock);
      PN->addIncoming(U->Update, OrigBlock);
    }
  }

  return true;
}



