
#include "llvm/GlobalVariable.h"
#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Support/Debug.h"

#include "Criticality.h"
#include "FaultInjectionUtils.h"
#include "InstructionInjection.h"
#include "LoopUpdateVars.h"
#include "Snapshot.h"

#include <set>
#include <vector>
#include <stack>

using namespace llvm;
using namespace vertical;


namespace {
  class LoopReuseInjection : public InstructionInjection {
  private:
    bool runOnFunction(Function &F);
    void getAnalysisUsage(AnalysisUsage &AU) const;
    bool isInstrumented(Instruction *I);

  public:
    static char ID;
    LoopReuseInjection() : InstructionInjection((intptr_t)&ID) {}

  private:
    std::set<Instruction *> AlreadyHandled;
  };
}


char LoopReuseInjection::ID = 0;
static RegisterPass<LoopReuseInjection> X(
    "loop-reuse", "Loop-ReUse Fault Injection");


void LoopReuseInjection::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequired<LoopUpdateVars>();
  AU.addRequired<Criticality>();
  AU.addRequired<Snapshot>();
  InstructionInjection::getAnalysisUsage(AU);
}


bool LoopReuseInjection::isInstrumented(Instruction *I) {
  switch (I->getOpcode()) {

    // memory creation (ADDED!)
    case Instruction::Malloc:
    // call instructions (ADDED!)
    case Instruction::Call:
    // ptr computation/access
    case Instruction::GetElementPtr:
    case Instruction::Load:
    case Instruction::Store:
    // comparison
    case Instruction::ICmp:
    case Instruction::FCmp:
    // control
    case Instruction::Select:
    // arithmetic/logic
    case Instruction::Add:
    case Instruction::Sub:
    case Instruction::Mul:
    case Instruction::UDiv:
    case Instruction::SDiv:
    case Instruction::FDiv:
    case Instruction::URem:
    case Instruction::SRem:
    case Instruction::FRem:
    case Instruction::Shl:
    case Instruction::LShr:
    case Instruction::AShr:
    case Instruction::And:
    case Instruction::Or:
    case Instruction::Xor:
    // casts
    case Instruction::FPToUI:
    case Instruction::FPToSI:
    case Instruction::UIToFP:
    case Instruction::SIToFP:
    // casts (ADDED!)
    case Instruction::Trunc:
    case Instruction::ZExt:
    case Instruction::SExt:
    case Instruction::FPTrunc:
    case Instruction::FPExt:
    case Instruction::PtrToInt:
    case Instruction::IntToPtr:
    case Instruction::BitCast:
      // supported
      return true;

    default:
      // unsupported
      return false;
  }
}


bool LoopReuseInjection::runOnFunction(Function &F) {

  Criticality    &CRT = getAnalysis<Criticality>();
  LoopUpdateVars &LUV = getAnalysis<LoopUpdateVars>();
  Snapshot       &SS  = getAnalysis<Snapshot>();

  Constant *Off = ConstantInt::get(Type::Int1Ty, 0);
  Constant *On  = ConstantInt::get(Type::Int1Ty, 1);

  typedef std::pair<UpdateInfo *, Instruction *> UpdateAndOperand;
  std::vector<UpdateAndOperand> QualifyingUpdates;

  for (LoopUpdateVars::iterator UI = LUV.begin(); UI != LUV.end(); ++UI) {

    UpdateInfo *U = *UI;

    Instruction *UpdateComp;
    if (StoreInst *S = dyn_cast<StoreInst>(U->Source)) {
      UpdateComp = cast<Instruction>(S->getOperand(0));
    } else {
      UpdateComp = U->Update;
    }

    if (UpdateComp->isBinaryOp()) {

      Instruction *Op = 0;
      if (UpdateComp->getOperand(0) == U->Source) {
        Op = dyn_cast<Instruction>(UpdateComp->getOperand(1));
      } else if (UpdateComp->getOperand(1) == U->Source) {
        Op = dyn_cast<Instruction>(UpdateComp->getOperand(0));
      }

      if (Op && isInstrumented(Op) /* might get a PHI */) {
        QualifyingUpdates.push_back(std::make_pair(U, Op));
      }
    }
  }

  forkInsts(F);

  for (std::vector<UpdateAndOperand>::iterator UI = QualifyingUpdates.begin();
       UI != QualifyingUpdates.end(); ++UI) {

    UpdateInfo  *U  = UI->first;
    Instruction *Op = UI->second;
    BasicBlock  *H  = U->Parent->getHeader();

    DOUT << "LoopReuseInjection: " << SS.IIMap[U->Update]->StringId << "\n";
    assert(InjectionMap.count(Op) != 0);

    if (CRT.CriticalityMap[Op] != 0) {
      // this instruction is critical -- don't do it!
      DOUT << "  Critical! Skipping....\n";
      continue;
    }

    if (!U->Parent->contains(Op->getParent())) {
      // the operand lies outside the loop -- skip it
      DOUT << "  Operand outside loop! Skipping....\n";
      continue;
    }

    if (AlreadyHandled.count(Op)) {
      // already done in earlier, inner loop
      DOUT << "  Already handled! Skipping....\n";
      continue; 
    }

    // we'll claim it
    AlreadyHandled.insert(Op);

    // insert the bypass bit variable
    BasicBlock::iterator InsertPos = 
      U->Update->getParent()->getParent()->begin()->begin();
    while (isa<PHINode>(InsertPos) || isa<AllocaInst>(InsertPos)) ++InsertPos;
    AllocaInst *ReuseBit = new AllocaInst(Type::Int1Ty, "ReuseLV", InsertPos); 

    // reset it at the beginning of each loop iteration
    new StoreInst(Off, ReuseBit, H->getFirstNonPHI());

    // iterate over the dependents of this function and mark them if they are
    // not critical
    for (UpdateInfo::iterator II = U->begin(), IE = U->end(); II != IE; ++II)
    {
      Instruction *OrigI = *II;

      if (InjectionMap.count(OrigI) == 0) {
        // this instruction is not instrumentable -- skip it
        DOUT << "  Skipping non-instrumentable instruction " <<
          SS.IIMap[OrigI]->StringId << "\n";

      } else if (CRT.CriticalityMap[OrigI] != 0) {
        // this instruction is critical -- skip it
        DOUT << "  Skipping critical instruction " <<
          SS.IIMap[OrigI]->StringId << "\n";

      } else {
        DOUT << "  Marking " << SS.IIMap[OrigI]->StringId << "\n";
        Instruction *FailI = InjectionMap[OrigI];
        new StoreInst(On, ReuseBit, FailI);
      }
    }

    // now implement the bypass
    InsertPos = H->begin();
    PHINode *ReusePHI = PHINode::Create(Op->getType(), "Reuse", InsertPos);

    typedef GraphTraits<Inverse<BasicBlock *> > InvBlockTraits;
    for (InvBlockTraits::ChildIteratorType PI = InvBlockTraits::child_begin(H);
         PI != InvBlockTraits::child_end(H); ++PI) {

      ReusePHI->addIncoming(Constant::getNullValue(Op->getType()), *PI);
    }

    DOUT << "  Creating unconditional reuse for failure\n";
    // for the Op instruction, first implement the failure behavior
    // (unconditional reuse)
    Instruction *FailReuse = InjectionMap[Op];
    FailReuse->replaceAllUsesWith(ReusePHI);

    DOUT << "  Creating conditional reuse for update\n";
    // now implement the non-failure reuse (conditional reuse)

    InsertPos = BasicBlock::iterator(Op);
    Value *IsReuse = new LoadInst(ReuseBit, "ReuseBit", ++InsertPos);
    Instruction *NewReuse = SelectInst::Create(IsReuse, ReusePHI,
                                               ReusePHI, // see setOperand below
                                               "NewReuse", InsertPos);
    Op->replaceAllUsesWith(NewReuse);
    NewReuse->setOperand(2, Op); // need to do this after replace all uses

    DOUT << "  Building dominator tree\n";
    // first we do a forward flow analysis to discover the reach of NewReuse
    std::stack<BasicBlock *> Successors;
    Successors.push(NewReuse->getParent());

    std::set<BasicBlock *> ReachedBlocks;
    while (!Successors.empty()) {
      BasicBlock *Current = Successors.top(); Successors.pop();
      ReachedBlocks.insert(Current);

      typedef GraphTraits<BasicBlock *> BlockTraits;
      for (BlockTraits::ChildIteratorType SI = BlockTraits::child_begin(Current);
           SI != BlockTraits::child_end(Current); ++SI) {

        BasicBlock *Next = *SI;
        if (U->Parent->contains(Next) && Next != H &&
            ReachedBlocks.count(Next) == 0) {
          Successors.push(Next);
        }
      }
    }

    // now check for PHIs; if there is no PHI where one is necessary, insert one
    std::set<BasicBlock *> VisitedSet;
    std::map<BasicBlock *, Instruction*> Versions;
    std::map<BasicBlock *, PHINode*>  NewPHIBlocks;

    typedef GraphTraits<BasicBlock *> BlockTraits;
    for (BlockTraits::ChildIteratorType SI =
         BlockTraits::child_begin(NewReuse->getParent()); SI !=
         BlockTraits::child_end(NewReuse->getParent()); ++SI) {
      Successors.push(*SI);
      Versions[*SI] = NewReuse;
    }

    while (!Successors.empty()) {
      BasicBlock  *CurrentBB = Successors.top(); Successors.pop();
      Instruction *CurrentI  = Versions[CurrentBB];

      if (VisitedSet.count(CurrentBB)) 
        // already visited
        continue;

      VisitedSet.insert(CurrentBB);

      bool PHINeeded = false;
      typedef GraphTraits<Inverse<BasicBlock *> > InvBlockTraits;
      for (InvBlockTraits::ChildIteratorType PI =
           InvBlockTraits::child_begin(CurrentBB); PI !=
           InvBlockTraits::child_end(CurrentBB); ++PI) {

        if (ReachedBlocks.count(*PI) == 0) {
          PHINeeded = true;
          break;
        }
      }

      if (PHINeeded) {
        // try and find one if PHI is needed
        PHINode *PN;
        DOUT << "  Checking for PHI\n";
        for (BasicBlock::iterator I = CurrentBB->begin();
             (PN = dyn_cast<PHINode>(I)) && PHINeeded; ++I) {

          for (unsigned Idx = 0; Idx != PN->getNumIncomingValues(); ++Idx) {
            if (PN->getIncomingValue(Idx) == CurrentI) {
              CurrentI = PN;
              PHINeeded = false;
              break;
            }
          }
        }

        if (PHINeeded) {
          // not found; insert one
          DOUT << "  PHI not found... inserting in " <<
            getStringId(CurrentBB) << "\n";
          PN = PHINode::Create(NewReuse->getType(), "ReusePHI",
                               CurrentBB->begin());

          NewPHIBlocks[CurrentBB] = PN;
          CurrentI = PN;
        }
      }

      typedef GraphTraits<BasicBlock *> BlockTraits;
      for (BlockTraits::ChildIteratorType SI =
           BlockTraits::child_begin(CurrentBB); SI !=
           BlockTraits::child_end(CurrentBB); ++SI) {

        BasicBlock *NextBB = *SI;
        if (NextBB == H) {
          unsigned Idx = ReusePHI->getBasicBlockIndex(CurrentBB);
          ReusePHI->setIncomingValue(Idx, CurrentI);

        } else if (U->Parent->contains(NextBB)) {
          Versions[NextBB] = CurrentI;
          Successors.push(NextBB);
        }
      }
    }

    for (std::map<BasicBlock *, PHINode *>::iterator PI = NewPHIBlocks.begin();
         PI != NewPHIBlocks.end(); ++PI) {

      BasicBlock *BB = PI->first;
      PHINode    *PN = PI->second;

      typedef GraphTraits<Inverse<BasicBlock *> > InvBlockTraits;
      for (InvBlockTraits::ChildIteratorType PI =
           InvBlockTraits::child_begin(BB); PI !=
           InvBlockTraits::child_end(BB); ++PI) {

        DOUT << "  adding incoming to PHI in " << getStringId(BB) << "\n";
        if (ReachedBlocks.count(*PI) == 0) {
          PN->addIncoming(ReusePHI, *PI);
        } else {
          PN->addIncoming(Versions[*PI], *PI);
        }
      }
    }
  }

  return true;
}


