
#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Support/Debug.h"
#include "llvm/Analysis/LoopInfo.h"

#include <stack>
#include <set>

#include "FaultInjectionUtils.h"
#include "LoopUpdateVars.h"

using namespace llvm;
using namespace vertical;


char LoopUpdateVars::ID = 0;
static RegisterPass<LoopUpdateVars> X(
    "loop-update-vars", "Loop Update Variable Analysis", true, true);

void LoopUpdateVars::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequiredTransitive<LoopInfo>();
  AU.setPreservesAll();
}


void LoopUpdateVars::registerUpdate(Instruction *Update, Instruction *Source,
                                    Loop *L) {

  UpdateInfo *U = new UpdateInfo();
  U->Update = Update;
  U->Source = Source;
  U->Parent = L;
  UpdateList.push_back(U);

  BasicBlock *H = L->getHeader();

  std::set<Instruction *> PHISet; // to check for cycles
  std::stack<Instruction *> DependentStack;
  DependentStack.push(Update);

  while (!DependentStack.empty()) {
    Instruction *I = DependentStack.top(); DependentStack.pop();

    if (isa<PHINode>(I)) {

      if (PHISet.count(I) == 0 && I->getParent() != H) {
        PHISet.insert(I); 
      } else {
        continue; // skip this one
      }
    }

    for (User::op_iterator Op = I->op_begin(); Op != I->op_end();
         ++Op) {

      Instruction *Dep = dyn_cast<Instruction>(Op->get());
      if (Dep && L->contains(Dep->getParent())) {

        U->Dependents.insert(Dep);
        DependentStack.push(Dep);
      }
    }
  }
}

void LoopUpdateVars::analyzeLoopAndSubloops(Loop *L) {

  for (Loop::iterator LI = L->begin(), LE = L->end(); LI != LE; ++LI) {
    // recurse on subloops
    analyzeLoopAndSubloops(*LI);
  }

  BasicBlock *H = L->getHeader();
  DOUT << "LoopUpdateVars: " << getStringId(H) << "\n";

  std::vector<BasicBlock *> BackEdges;

  typedef GraphTraits<Inverse<BasicBlock *> > InvBlockTraits;
  for (InvBlockTraits::ChildIteratorType PI = InvBlockTraits::child_begin(H);
       PI != InvBlockTraits::child_end(H); ++PI) {

    BasicBlock *BB = *PI;
    if (L->contains(BB)) {
      BackEdges.push_back(BB);
    }
  }

  // If there are loads in the loop, check for a matching store.  If so, we
  // have a candidate
  for (Loop::block_iterator BI = L->block_begin(), BIE = L->block_end();
       BI != BIE; ++BI) {
    for (BasicBlock::iterator II = (*BI)->begin(), IIE = (*BI)->end(); II !=
         IIE; ++II) {
      if (LoadInst *Ld = dyn_cast<LoadInst>(II)) {

        Value *Ptr = Ld->getPointerOperand();
        for (Value::use_iterator UI = Ptr->use_begin(), UE = Ptr->use_end();
             UI != UE; ++UI) {

          if (StoreInst *S = dyn_cast<StoreInst>(*UI)) {
            if (L->contains(S->getParent())) {
              registerUpdate(S, Ld, L);
            }
          }
        }
      }
    }
  }

  // Loop over all of the PHI nodes, looking for an internally updated
  // variable.
  for (BasicBlock::iterator II = H->begin(); isa<PHINode>(II);
       ++II) {

    PHINode *PN = cast<PHINode>(II);
    for (std::vector<BasicBlock *>::iterator BEI = BackEdges.begin(),
         BEIE = BackEdges.end(); BEI != BEIE; ++BEI) {

      if (Instruction *I =
          dyn_cast<Instruction>(PN->getIncomingValueForBlock(*BEI))) {

        std::set<Instruction *> PHISet; // to check for cycles
        std::stack<Instruction *> CandidateStack;
        CandidateStack.push(I);

        while (!CandidateStack.empty()) {

          I = CandidateStack.top(); CandidateStack.pop();

          // if I is a PHINode, we search backward until we find a non-PHI
          // (if any)
          if (isa<PHINode>(I)) {

            if (PHISet.count(I) == 0 && I->getParent() != H) {
              PHISet.insert(I);

              // keep looking (recurse)
              for (User::op_iterator Op = I->op_begin(); Op != I->op_end();
                   ++Op) {

                Instruction *Dep = dyn_cast<Instruction>(Op->get());
                if (Dep && L->contains(Dep->getParent())) {
                  CandidateStack.push(Dep);
                }
              }
            }

          } else {
            // found a non-PHI (the latest update)
            registerUpdate(I, PN, L);
          }
        }
      }
    }
  }
}

void LoopUpdateVars::releaseMemory() {
  for (iterator UI = begin(); UI != end(); ++UI)
    delete *UI;

  UpdateList.clear();
}

bool LoopUpdateVars::runOnFunction(Function &F) {

  releaseMemory();

  LoopInfo &LIN = getAnalysis<LoopInfo>();
  for (LoopInfo::iterator LI = LIN.begin(), LE = LIN.end(); LI != LE; ++LI) {
    // this will analyze the loops in order of innermost to outermost
    analyzeLoopAndSubloops(*LI);
  }

  return false;
}


void LoopUpdateVars::print(std::ostream &O, const Module *M) const {

  for (const_iterator UI = begin(), UE = end(); UI != UE; ++UI) {

    UpdateInfo *U = (*UI);
    O << "Src: " << getStringId(U->Update) << "\n";

    for (UpdateInfo::iterator II = U->begin(), IE = U->end(); II != IE; ++II) {
      O << "  Dep: " << getStringId(*II) << "\n";
    }
  }
}



