
#include "llvm/Constants.h"
#include "llvm/GlobalVariable.h"
#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Analysis/Dominators.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/CFG.h"

#include "InstructionInjection.h"
#include "RealTranslate.h"
#include "Snapshot.h"

#include <iostream>

#include <set>
#include <vector>
#include <stack>

using namespace llvm;
using namespace vertical;


namespace {

  class RealInjection : public InstructionInjection {
  private:
    bool runOnFunction(Function &F);
    /* FIXME: */ bool isInstrumented(Instruction *I) { return false; };
    void getAnalysisUsage(AnalysisUsage &AU) const;

  public:

    static char ID;
    RealInjection() : InstructionInjection((intptr_t)&ID) {}
  };
}


char RealInjection::ID = 0;
static RegisterPass<RealInjection> X(
    "real-inject", "Real Fault Injection");


void RealInjection::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequired<DominanceFrontier>();
  AU.addRequired<DominatorTree>();
  AU.addRequired<LoopInfo>();
  AU.addRequired<RealTranslate>();
  AU.addRequired<Snapshot>();
  InstructionInjection::getAnalysisUsage(AU);
}

bool RealInjection::runOnFunction(Function &F) {

  DominanceFrontier &DF = getAnalysis<DominanceFrontier>();
  DominatorTree &DT     = getAnalysis<DominatorTree>();
  LoopInfo &LI          = getAnalysis<LoopInfo>();
  RealTranslate &RT     = getAnalysis<RealTranslate>();
  Snapshot &SS          = getAnalysis<Snapshot>();

  std::map<StoreInst *, PHINode *> DefToPHIMap;
  std::map<Instruction *, Instruction *> InjectMap;

  DOUT << "RealInjection: " << F.getName() << "\n";

  // insert PHIs now, we will need them later
  for (std::set<Region *>::iterator RI = RT.Regions.begin(),
       RE = RT.Regions.end(); RI != RE; ++RI) {
    Region *RR = *RI;

    // for each escape definition insert phi nodes in the catch and end
    // regions 
    for (std::map<GlobalVariable *, StoreInst *>::iterator PI =
         RR->EscapeDefs.begin(), PE = RR->EscapeDefs.end();
         PI != PE; ++PI) {

      Instruction *Def = cast<Instruction>(PI->second->getOperand(0));

      PHINode *CatchPHI = 0;
      if (RR->Catch) {
        // we need to create a PHI at the start of the catch region.
        // we will deal with the CatchPHI after we have forked the instructions,
        // so we save it away for now
        DOUT << "  Creating CatchPHI\n";
        CatchPHI = PHINode::Create(
            Def->getType(), "CatchPHI", RR->Catch->begin());
      }

      PHINode *EndPHI = 0;
      // Check to see if there is a PHI for this definition in the end
      // region.  If so, we don't need to create one.
      std::set<Instruction *> SeenSet;
      std::stack<Instruction *> DefUsers;
      DefUsers.push(Def);
      
      while (!DefUsers.empty()) {
        Instruction *DF = DefUsers.top(); DefUsers.pop();
        SeenSet.insert(DF);

        for (Value::use_iterator UI = DF->use_begin(), UE = DF->use_end();
            UI != UE; ++UI) {

          PHINode *PHI = dyn_cast<PHINode>(*UI);
          if (PHI && !SeenSet.count(PHI)) {
            if (PHI->getParent() == RR->End) {
              DOUT << "  Appropriating EndPHI\n";
              EndPHI = PHI;
              goto FoundEndPHI;

            } else {
              DefUsers.push(PHI);
            }
          }
        }
      }

FoundEndPHI:

      // not found; create it
      if (!EndPHI) {
        DOUT << "  Creating EndPHI\n";
        EndPHI = PHINode::Create(Def->getType(), "EndPHI", RR->End->begin());

        // change the uses of Def after the region end to use the EndPHI
        std::vector<Use *> UseList;
        for (Value::use_iterator UI = Def->use_begin(), UE = Def->use_end();
             UI != UE; ++UI) {
          UseList.push_back(&UI.getUse());
        }

        DominanceFrontier::iterator DefFrontier = DF.find(Def->getParent());
        Loop *EndPHILoop = LI.getLoopFor(EndPHI->getParent());

        assert(DefFrontier != DF.end());
        
        for (std::vector<Use *>::iterator UI = UseList.begin(), UE =
             UseList.end(); UI != UE; ++UI) {
          Instruction *U = cast<Instruction>((*UI)->getUser());

          PHINode *PN = dyn_cast<PHINode>(U);
          BasicBlock *UBB = 0; /* the use basic block */

          if (PN) {
            for (unsigned i = 0; i < PN->getNumIncomingValues(); ++i) {
              if (PN->getIncomingValue(i) == Def) {
                UBB = PN->getIncomingBlock(i);
                break;
              }
            }
          } else {
            UBB = U->getParent();
          }
          assert(UBB);

          if (U && (DT.dominates(EndPHI->getParent(), UBB) ||
                    (EndPHILoop && 
                     UBB == EndPHILoop->getHeader()))) {
                    // TODO: this is intended to detect loops... not sure 
                    // if it's 100% correct
                    // (DefFrontier->second.count(U->getParent()) &&
                    //  DT.dominates(U, Def)))) 
            (*UI)->set(EndPHI);
          }
        }

        // add necessary predecessors to PHI
        for (pred_iterator PI = pred_begin(RR->End), PE = pred_end(RR->End); 
             PI != PE; ++PI) {
          BasicBlock *P = *PI;

          DOUT << "    Adding entry for " << P->getName() << "\n";

          // order matters here
          if (RR->Catch && DT.dominates(RR->Catch, P)) { 
            // CatchPHI coming from the region catch
            EndPHI->addIncoming(CatchPHI, P);
          } else if (DT.dominates(RR->Try, P)) { 
            // Def coming from the try region
            EndPHI->addIncoming(Def, P);
          } else {
            assert(0);
          }
        }
      }

      if (EndPHI->getNumIncomingValues() == 0) {
        EndPHI->eraseFromParent();
        EndPHI = 0;
      }
      
      // where is the jump-on-failure destination?
      if (!RR->Catch) {
        assert(EndPHI);
        DefToPHIMap[PI->second] = EndPHI;
      } else {
        DefToPHIMap[PI->second] = CatchPHI;
      }
    }
  }

  // accumulate the injection sets into a single set for fork
  std::vector<Instruction *> InjectList;
  for (std::set<Region *>::iterator RI = RT.Regions.begin(),
       RE = RT.Regions.end(); RI != RE; ++RI) {
    Region *RR = *RI;
    for (std::vector<Instruction *>::iterator II = RR->InjectList.begin(),
         IE = RR->InjectList.end(); II != IE; ++II) {
      InjectList.push_back(*II);
    }
  }

  // fork the instructions in the inject list
  forkInsts(F, InjectList, InjectMap);

  // reset the global variables associated with each escape variable
  BasicBlock::iterator InsertPos = F.getEntryBlock().begin();
  while (isa<AllocaInst>(InsertPos) || isa<PHINode>(InsertPos)) 
    ++InsertPos;

  for (std::map<StoreInst *, GlobalVariable *>::iterator PI =
       RT.EscapeDefs.begin(), PE = RT.EscapeDefs.end(); PI != PE; ++PI) {
    GlobalVariable *Val = PI->second;
    new StoreInst(Val->getInitializer(), Val, InsertPos);
  }

  // Change the failure paths so that they branch to the region catch, or end
  // if there is no catch.  The catch PHI nodes need to be updated to accept the
  // correct instance of the escape variable for each failed instruction.
  for (std::set<Region *>::iterator RI = RT.Regions.begin(),
       RE = RT.Regions.end(); RI != RE; ++RI) {
    Region *RR = *RI;

    DOUT << "  Region " << SS.BIMap[RR->Try]->StringId << "\n";

    for (std::vector<Instruction *>::iterator II = RR->InjectList.begin(),
         IE = RR->InjectList.end(); II != IE; ++II) {

      Instruction *Orig = *II;
      Instruction *Fail = InjectMap[Orig];

      DOUT << "  Redirecting " << SS.IIMap[Orig]->StringId << "\n";

      BasicBlock *FailDest = RR->Catch;
      if (!RR->Catch)
        FailDest = RR->End;

      TerminatorInst *FailTerm = Fail->getParent()->getTerminator();
      BranchInst *FailBranch = dyn_cast<BranchInst>(FailTerm);

      if (!FailBranch) {
        FailBranch = BranchInst::Create(FailDest, FailTerm);
        FailTerm->eraseFromParent();
        // TODO: update PHIs if switch or other non-returning inst

      } else {
        // change the output path of the failed instruction to FailDest
        FailBranch->setUnconditionalDest(FailDest);
      }

      // add incoming values to the region catch/end for each escape definition
      // inside the try region
      for (std::map<GlobalVariable *, StoreInst *>::iterator PI =
          RR->EscapeDefs.begin(), PE = RR->EscapeDefs.end(); PI != PE; ++PI) {

        GlobalVariable *Val = PI->first;
        StoreInst      *Def = PI->second;
        PHINode        *LastPHI = DefToPHIMap[Def];

        // in the case of failure, we load the global variable and pass that
        // forward
        LoadInst *PLoad = new LoadInst(Val, "EscapeLoad", FailBranch);
        LastPHI->addIncoming(PLoad, Fail->getParent());
      }

      if (!isa<TerminatorInst>(Orig)) {

        // remove the Fail instruction and its use (if any) because it is no
        // longer valid (in the case of a terminator instruction we have already
        // effectively erased it)
        assert(Fail->use_empty() || Fail->hasOneUse());
        if (Fail->hasOneUse()) {
          PHINode *UsePHI = cast<PHINode>(*(Fail->use_begin()));
          UsePHI->removeIncomingValue(Fail->getParent());
        }
        Fail->eraseFromParent();
      }
    }
  }

  // this is for counting the number of REAL block boundary crossings
  Constant *RegisterFn =
    F.getParent()->getOrInsertFunction("__POKE__RegisterReal",
                                       Type::VoidTy, (Type *)0);

  Constant *RegisterFn2 =
    F.getParent()->getOrInsertFunction("__POKE__RegisterRealRecover",
                                       Type::VoidTy, (Type *)0);

  for (std::set<Region *>::iterator RI = RT.Regions.begin(),
       RE = RT.Regions.end(); RI != RE; ++RI) {
    Region *RR = *RI;

    assert(RR->Try != &F.getEntryBlock());
    CallInst::Create(RegisterFn, "", RR->Try->getFirstNonPHI());

    if (RR->Catch)
      CallInst::Create(RegisterFn2, "", RR->Catch->getFirstNonPHI());
  }

  return true;
}



