
#include "llvm/Constants.h"
#include "llvm/GlobalVariable.h"
#include "llvm/Function.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Analysis/Dominators.h"
#include "llvm/Analysis/LoopInfo.h"
#include "llvm/Support/Debug.h"
#include "llvm/Support/CFG.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"

#include "RealTranslate.h"
#include "FaultInjectionUtils.h"

#include <sstream>
#include <set>
#include <map>
#include <stack>

using namespace llvm;
using namespace vertical;

char RealTranslate::ID = 0;
static RegisterPass<RealTranslate> X(
    "real-translate", "Real Translation");

void RealTranslate::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequired<DominatorTree>();
  AU.addRequired<LoopInfo>();
}

bool RealTranslate::runOnFunction(Function &F) {

  Module &M = *F.getParent();

  EscapeDefs.clear();
  for (std::set<Region *>::iterator RI = Regions.begin(),
       RE = Regions.end(); RI != RE; ++RI) {
    delete (*RI);
  }
  Regions.clear();

  DT = &getAnalysis<DominatorTree>();

  std::map<unsigned, Region *> ID2RegionMap;
  std::map<unsigned, GlobalVariable *> ID2ValMap;

  // first find all marker functions and extract appropriate data and/or create
  // appropriate structures
  for (Function::iterator BB = F.begin(), BBE = F.end(); BB != BBE; ++BB) {

    BasicBlock::iterator I = BB->begin();
    BasicBlock::iterator IE = BB->end();

    bool SeenLastAtomic = false;
   
    while (I != IE) {

      CallInst *CI = dyn_cast<CallInst>(I);
      ++I; // advance iterator so we can erase CI later

      if (CI && CI->getCalledFunction()) {
        std::string FuncName = CI->getCalledFunction()->getName();

        if (FuncName.find("llvm_real_escape_declare") == 0) {
          // Operand 1 = numeric id
          // Operand 2 = initializer

          // read the id
          std::stringstream IDString;
          unsigned ID = cast<ConstantInt>(
              CI->getOperand(1))->getValue().getZExtValue();
          IDString << "Val" << ID;

          // create the global variable
          Constant *Initializer = cast<Constant>(CI->getOperand(2));
          GlobalVariable *Val = new GlobalVariable(
              Initializer->getType(), false /* not constant */,
              GlobalValue::InternalLinkage, Initializer, IDString.str(),
              &M);

          // store the global variable
          ID2ValMap[ID] = Val;

          // replace all uses with the initializer; we need to store the uses in
          // a separate list so that we don't affect the use list itself while
          // we are iterating
          std::vector<Use *> UseList;
          for (Value::use_iterator UI = CI->use_begin(), UE = CI->use_end();
               UI != UE; ++UI) {
            UseList.push_back(&UI.getUse());
          }
          for (std::vector<Use *>::iterator UI = UseList.begin(), UE =
               UseList.end(); UI != UE; ++UI) {
            (*UI)->set(Initializer);
          }
          CI->eraseFromParent();

        } else if (FuncName.find("llvm_real_escape_assign") == 0) {
          // Operand 1 = numeric id
          // Operand 2 = defining instruction 

          // read the id and defining instruction
          unsigned ID = cast<ConstantInt>(
              CI->getOperand(1))->getValue().getZExtValue();
          Instruction *Def = cast<Instruction>(CI->getOperand(2));

          assert(ID2ValMap.count(ID));
          GlobalVariable *GVal = ID2ValMap[ID];

          // associate the definition with its global variable
          // and create the associated store instructions
          if (FuncName.find("llvm_real_escape_assign_atomic") == 0) {
            assert(!SeenLastAtomic);
            EscapeDefs[new StoreInst(Def, GVal)] = GVal;

            if (FuncName.find("llvm_real_escape_assign_atomic_last") == 0) {
              for (std::map<StoreInst *, GlobalVariable *>::iterator PI =
                   EscapeDefs.begin(), PE = EscapeDefs.end(); PI != PE; ++PI) {

                StoreInst *Def = PI->first;
                Def->insertBefore(CI);
                SeenLastAtomic = true;
              }
            }
          } else {
            EscapeDefs[new StoreInst(Def, GVal, CI)] = GVal;
          }

          // replace all uses with the incoming definition
          assert(!CI->use_empty() && "escape assignment not used!");
          CI->replaceAllUsesWith(Def);
          CI->eraseFromParent();

        } else if (FuncName.find("llvm_real_try") == 0) {
          // Operand 1 = numeric id

          unsigned ID = cast<ConstantInt>(
              CI->getOperand(1))->getValue().getZExtValue();
          if (ID2RegionMap.count(ID) == 0) {
            Region *RR = new Region();
            ID2RegionMap[ID] = RR;
            Regions.insert(RR);
          }

          std::stringstream TryString;
          TryString << "Try" << ID;

          // create the header block
          BasicBlock *Header = CI->getParent();
          BasicBlock *Try = SplitBlock(Header, CI, this);

          I = Try->begin(); ++I;
          IE = Try->end();

          // record this basic block
          ID2RegionMap[ID]->Header = Header;
          ID2RegionMap[ID]->Try = Try;
          CI->eraseFromParent();

        } else if (FuncName.find("llvm_real_split") == 0) {
          // The purpose of this call is only to prevent dead code elimination
          // of code that is not in fact dead code

          LoadInst   *LD  = cast<LoadInst>(I++);
          CmpInst    *CMP = cast<CmpInst>(I++);
          BranchInst *BR  = cast<BranchInst>(I++);

          // we will remove link the link the first successor
          // keep the link to the second
          // TODO: assert that this is the right BB
          RemoveSuccessor(BR, 0);

          // erase in reverse order to prevent def-use issues
          CMP->eraseFromParent();
          LD->eraseFromParent();
          CI->eraseFromParent();

        } else if (FuncName.find("llvm_real_catch") == 0) {
          // Operand 1 = numeric id

          unsigned ID = cast<ConstantInt>(
              CI->getOperand(1))->getValue().getZExtValue();
          if (ID2RegionMap.count(ID) == 0) {
            Region *RR = new Region();
            ID2RegionMap[ID] = RR;
            Regions.insert(RR);
          }

          // record this basic block
          ID2RegionMap[ID]->Catch = CI->getParent();
          CI->eraseFromParent();

        } else if (FuncName.find("llvm_real_end") == 0) {
          // Operand 1 = numeric id

          unsigned ID = cast<ConstantInt>(
              CI->getOperand(1))->getValue().getZExtValue();
          if (ID2RegionMap.count(ID) == 0) {
            Region *RR = new Region();
            ID2RegionMap[ID] = RR;
            Regions.insert(RR);
          }

          // record this basic block
          ID2RegionMap[ID]->End = CI->getParent();
          CI->eraseFromParent();
        }
      }
    }
  }

  // detect null catch regions
  for (std::set<Region *>::iterator RI = Regions.begin(),
       RE = Regions.end(); RI != RE; ++RI) {
    Region *RR = *RI;

    BranchInst *BR = dyn_cast<BranchInst>(RR->Catch->getTerminator());

    // if this basic block is essentially empty and branches directly to the end
    // block, mark the region catch as NULL (0) and break out of this basic
    // block
    if (BR && BR->isUnconditional() &&
        BR->getSuccessor(0) == RR->End &&
        RR->Catch->size() == 2) {

      RemoveSuccessor(BR, 0);
      RR->Catch = 0;
    }
  }

  // associate each escape definition with its try region
  for (std::map<StoreInst *, GlobalVariable *>::iterator
       PI = EscapeDefs.begin(), PE = EscapeDefs.end(); PI != PE; ++PI) {

    StoreInst      *Def = PI->first;
    GlobalVariable *Val = PI->second;

    bool found = false;
    for (std::set<Region *>::iterator RI = Regions.begin(),
        RE = Regions.end(); RI != RE; ++RI) {
      Region *RR = *RI;

      if (DT->dominates(RR->Try, Def->getParent()) &&
          !DT->dominates(RR->End, Def->getParent())) {
        assert(RR->EscapeDefs.count(Val) == 0 && 
                "escape variables should only be defined once in a region");

        RR->EscapeDefs[Val] = Def;
        found = true;
        break;
      }
    }

    assert(found && "escape variable definition not inside a region");
  }

  DOUT << "RealTranslate: " << F.getName() << "\n";

  // register the injections on each region independently
  for (std::set<Region *>::iterator RI = Regions.begin(),
       RE = Regions.end(); RI != RE; ++RI) {
    Region *RR = *RI;

    std::set<BasicBlock *>   BBVisited;
    std::stack<BasicBlock *> BBToVisit;

    DOUT << " Header: "<< RR->Header->getName()   << "\n";
    DOUT << " Try: " << RR->Try->getName()   << "\n";
    if (RR->Catch) 
      DOUT << " Catch: " << RR->Catch->getName() << "\n";
    DOUT << " End: " << RR->End->getName()   << "\n";

    // move instructions that violate the semantics of the try region to the
    // header region
    BBToVisit.push(RR->Try);
    BBVisited.insert(RR->Try);
    while (!BBToVisit.empty()) {

      BasicBlock *BB = BBToVisit.top(); BBToVisit.pop();

      BasicBlock::iterator II = BB->getFirstNonPHI();
      BasicBlock::iterator IE = BB->getTerminator();

      for (succ_iterator SI = succ_begin(BB), SE = succ_end(BB); 
           SI != SE; ++SI) {
        BasicBlock *S = *SI;

        // ensure the successor block is in the try region before adding it to
        // the stack
        if (!DT->dominates(RR->End,   S) &&
            (!RR->Catch || !DT->dominates(RR->Catch, S)) &&
            !BBVisited.count(S)) {
          BBToVisit.push(S);
          BBVisited.insert(S);
        }
      }
    }

    // now iterate over all blocks in the try region and mark all instrumented
    // instructions for injection
    BBVisited.clear();
    BBToVisit.push(RR->Try);
    BBVisited.insert(RR->Try);
    while (!BBToVisit.empty()) {

      BasicBlock *BB = BBToVisit.top(); BBToVisit.pop();

      for (BasicBlock::iterator I = BB->getFirstNonPHI(),
           IE = BB->end(); I != IE; ++I) {
        // currently only branch and return terminating instructions are
        // supported
        assert(!isa<TerminatorInst>(I) ||
               isa<BranchInst>(I) ||
               isa<ReturnInst>(I));

        StoreInst *S = dyn_cast<StoreInst>(I);
        if (!S || !EscapeDefs.count(S)) {
          RR->InjectList.push_back(I);
        }
      }

      for (succ_iterator SI = succ_begin(BB), SE = succ_end(BB); 
           SI != SE; ++SI) {
        BasicBlock *S = *SI;

        // ensure the successor block is in the try region before adding it to
        // the stack
        if (!DT->dominates(RR->End,   S) &&
            (!RR->Catch || !DT->dominates(RR->Catch, S)) &&
            !BBVisited.count(S)) {
          BBToVisit.push(S);
          BBVisited.insert(S);
        }
      }
    }
  }


  std::set<BasicBlock *> TryBlocks;
  for (std::set<Region *>::iterator RI = Regions.begin(),
       RE = Regions.end(); RI != RE; ++RI) {
    Region *RR = *RI;
    TryBlocks.insert(RR->Try);
  }

  // clean-up
  for (std::set<Region *>::iterator RI = Regions.begin(),
       RE = Regions.end(); RI != RE; ++RI) {
    Region *RR = *RI;
    BasicBlock *SuccBB, *PrevBB;

    SuccBB = RR->Header;
    PrevBB = SuccBB->getSinglePredecessor();
    while (PrevBB &&
           !TryBlocks.count(PrevBB) &&
           MergeBlockIntoPredecessor(SuccBB, this)) {
      SuccBB = PrevBB;
      PrevBB = SuccBB->getSinglePredecessor();
      RR->Header = SuccBB;
    }

    SuccBB = RR->Try;
    PrevBB = SuccBB->getSinglePredecessor();
    while (PrevBB &&
           !TryBlocks.count(PrevBB) &&
           PrevBB != &PrevBB->getParent()->getEntryBlock() && 
           PrevBB->size() == 1 &&
           MergeBlockIntoPredecessor(SuccBB, this)) {
      TryBlocks.erase(SuccBB);
      TryBlocks.insert(PrevBB);
      SuccBB = PrevBB;
      PrevBB = SuccBB->getSinglePredecessor();
      RR->Try = SuccBB;
    }
  }

  return true;
}

void RealTranslate::print(std::ostream &O, const Module *M) const {

  for (std::set<Region *>::iterator RI = Regions.begin(),
       RE = Regions.end(); RI != RE; ++RI) {
    Region *RR = *RI;
    
    O << "Entry Block: " << getStringId(RR->Try) << "\n";

    for (std::vector<Instruction *>::iterator II = RR->InjectList.begin(),
        IE = RR->InjectList.end(); II != IE; ++II) {
      O << getStringId(*II) << "\n";
    }
  }
}


