//===- BlockProfiling.cpp - Insert counters for block profiling -----------===//
//
//                      The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This pass instruments the specified program with counters for basic block or
// function profiling.  This is the most basic form of profiling, which can tell
// which blocks are hot, but cannot reliably detect hot paths through the CFG.
// Block profiling counts the number of times each basic block executes, and
// function profiling counts the number of times each function is called.
//
// Note that this implementation is very naive.  Control equivalent regions of
// the CFG should not require duplicate counters, but we do put duplicate
// counters in.
//
//===----------------------------------------------------------------------===//

#include "llvm/Constants.h"
#include "llvm/Instructions.h"
#include "llvm/DerivedTypes.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Support/CommandLine.h"
#include "llvm/Support/Compiler.h"
#include "llvm/Support/Streams.h"
#include "llvm/Support/Debug.h"
#include "llvm/Transforms/Instrumentation.h"
#include "../lib/Transforms/Instrumentation/RSProfiling.h"
#include "../lib/Transforms/Instrumentation/ProfilingUtils.h"


using namespace llvm;


namespace {
  class BlockProfiler : public RSProfilers_std {
    bool runOnModule(Module &M);
  public:
    static char ID;

  private:
    void IncrementCounterInBlock(BasicBlock *BB, unsigned CounterNum, Value *CounterArray);
  };

  char BlockProfiler::ID = 0;
  RegisterPass<BlockProfiler> Y("block-profiling",
                                "Insert instrumentation for block profiling");
  RegisterAnalysisGroup<RSProfilers> YG(Y);
}


static cl::opt<bool> ProfilingHooks(
    "profiling-hooks", cl::Hidden, 
    cl::desc("program calls profiling hooks"));

static cl::opt<bool> CountLLVM(
    "count-llvm", cl::Hidden, 
    cl::desc("program counts llvm instructions"));

bool BlockProfiler::runOnModule(Module &M) {
  Function *MainFn = M.getFunction("main");
  if (MainFn == 0) {
    cerr << "WARNING: cannot insert block profiling into a module"
      << " with no main function!\n";
    return false;  // No main, no instrumentation!
  }

  if (CountLLVM) {

    // count the number of blocks
    unsigned NumBlocks = 0;
    for (Module::iterator F = M.begin(); F != M.end(); ++F) {
      NumBlocks += F->size();
    }

    DOUT << "Counted " << NumBlocks << " basic blocks\n";

    // declare the global counter array
    const Type *ATy = ArrayType::get(Type::Int64Ty, NumBlocks);
    GlobalVariable *Counters =
      new GlobalVariable(ATy, false, GlobalValue::InternalLinkage,
                        Constant::getNullValue(ATy), "BlockProfCounters", &M);

    Constant *ProfileFn = M.getOrInsertFunction(
        "__POKE__ProfileAccum",
        Type::VoidTy,
        Type::Int32Ty,
        PointerType::getUnqual(Type::Int32Ty),
        Type::Int32Ty,
        NULL);

    // Instrument all of the blocks...
    unsigned i = 0;
    for (Module::iterator F = M.begin(); F != M.end(); ++F) {

      if (F->begin() == F->end())
        continue;

      DOUT << "Instrumenting function " << F->getName() << "\n";

      BasicBlock *Entry = F->begin();
      BasicBlock::iterator InsertPos = Entry->begin();
      while (isa<AllocaInst>(InsertPos)) ++InsertPos;

      // Allocate the profile data array
      Constant *ArraySize = ConstantInt::get(Type::Int32Ty, F->size());
      AllocaInst *Array = new AllocaInst(
          Type::Int32Ty,
          ArraySize,
          "Array",
          InsertPos);

      // Set the array to all zeroes
      Constant *MemsetFn = M.getOrInsertFunction(
          "llvm.memset.i32", Type::VoidTy, PointerType::getUnqual(Type::Int8Ty),
          Type::Int8Ty, Type::Int32Ty, Type::Int32Ty, NULL);

      std::vector<Value*> MemsetArgs(4);
      MemsetArgs[0] = CastInst::Create(CastInst::BitCast, Array,
                                      PointerType::getUnqual(Type::Int8Ty),
                                      "CastToBytePtr", InsertPos);
      MemsetArgs[1] = ConstantInt::get(Type::Int8Ty, 0);
      MemsetArgs[2] = ConstantInt::get(Type::Int32Ty, F->size() * 4);
      MemsetArgs[3] = ConstantInt::get(Type::Int32Ty, 0);
      CallInst::Create(MemsetFn, MemsetArgs.begin(), MemsetArgs.end(), "",
                      InsertPos);

      unsigned FuncIndex = i;
      for (Function::iterator BB = F->begin(); BB != F->end(); ++BB) {
        DOUT << " Instrumenting block " << BB->getName() << "\n";

        // Insert counter at the start of the block
        IncrementCounterInBlock(BB, i - FuncIndex, Array);
        ++i;

        // For any return instructions, insert the retire call
        std::vector<Instruction*> retInsts;
        for (BasicBlock::iterator I = BB->begin(); I != BB->end(); ++I) {
          if (isa<ReturnInst>(I) || isa<UnwindInst>(I)) {
            retInsts.push_back(I);
          }
        }

        std::vector<Value*> ProfileArgs(3);
        ProfileArgs[0] = ConstantInt::get(Type::Int32Ty, FuncIndex);
        ProfileArgs[1] = Array;
        ProfileArgs[2] = ArraySize;
        for (unsigned j = 0; j < retInsts.size(); ++j) {
          DOUT << " Inserting call before return instruction\n";
          CallInst::Create(ProfileFn, ProfileArgs.begin(),
                          ProfileArgs.end(), "", retInsts[j]);
        }
      }
    }

    DOUT << "Instrumented " << i << " basic blocks \n";
    assert((i == NumBlocks) && "internal error!");

    // Add the initialization call to main.
    InsertProfilingInitCall(MainFn, "__POKE__ProfilingInit", Counters);
  }

  if (!ProfilingHooks) {
    // Add call if hooks not in program
    BasicBlock::iterator I = MainFn->begin()->begin();
    while (isa<AllocaInst>(I)) ++I;

    Constant *InitFn = M.getOrInsertFunction("__POKE__ProfilingAuto",
                                   Type::VoidTy, (Type *)0);
    CallInst::Create(InitFn, "", I);
  }

  return true;
}

void BlockProfiler::IncrementCounterInBlock(BasicBlock *BB, unsigned CounterNum,
                                            Value *CounterArray)
{
  Instruction *InsertPos = BB->getTerminator();

  Value *ElementPtr = GetElementPtrInst::Create(CounterArray,
                                                ConstantInt::get(Type::Int32Ty, CounterNum),
                                                "ArrayIndex", InsertPos);

  Value *OldVal = new LoadInst(ElementPtr, "OldCounter", InsertPos);
  profcode.insert(OldVal);
  Value *NewVal = BinaryOperator::CreateAdd(OldVal,
                                            ConstantInt::get(Type::Int32Ty, 1),
                                            "NewCounter", InsertPos);
  profcode.insert(NewVal);
  profcode.insert(new StoreInst(NewVal, ElementPtr, InsertPos));
}

