
#include "llvm/BasicBlock.h"
#include "llvm/Function.h"
#include "llvm/Instruction.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Value.h"
#include "llvm/ADT/BitVector.h"
#include "llvm/Analysis/Dominators.h"

#include "../FaultInjection/FaultInjectionUtils.h"

#include <stack>
#include <set>

using namespace vertical;
using namespace llvm;

class Relaxer : public FunctionPass {
private:
  bool runOnFunction(Function &F);
  void getAnalysisUsage(AnalysisUsage &AU) const;
  void print(std::ostream &O, const Module *M = 0) const;

  void MarkInterveningInsts(Instruction *start_inst,
                            Instruction *end_inst);

  DominatorTree *dom_tree;
  BitVector                         inst_vector;
  DenseMap<Instruction *, unsigned> inst_id_map;

public:
  static char ID;
  Relaxer() : FunctionPass((intptr_t)&ID) {}
};


char Relaxer::ID = 0;
static RegisterPass<Relaxer> X(
    "relaxer", "Relaxer Analysis", true, true);

void Relaxer::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequired<DominatorTree>();
  AU.setPreservesAll();
}

void Relaxer::MarkInterveningInsts(Instruction *start_inst,
                                   Instruction *end_inst) {

  BasicBlock *start_block = start_inst->getParent();
  BasicBlock *end_block   = end_inst->getParent();

  std::stack<BasicBlock *> blocks_to_visit;
  std::set<BasicBlock *>   blocks_visited;

  blocks_to_visit.push(end_block);
  blocks_visited.insert(end_block);

  while (!blocks_to_visit.empty()) {
    BasicBlock *block = blocks_to_visit.top(); blocks_to_visit.pop();

    BasicBlock::iterator inst_iterator;
    if (block == end_block) {
      inst_iterator = end_inst;

    } else {
      inst_iterator = &block->back();
    }

    // visit
    BasicBlock::iterator last = block->begin();
    BasicBlock::iterator term = start_inst;
    inst_vector.set(inst_id_map[inst_iterator]);
    while (inst_iterator != term && inst_iterator != last) {
      inst_iterator--;
      inst_vector.set(inst_id_map[inst_iterator]);
    } 

    // depth first
    if (block != start_block) {
      for (pred_iterator PI = pred_begin(block), PE = pred_end(block); 
           PI != PE; ++PI) {
        BasicBlock *pred_block = *PI;

        if (!blocks_visited.count(pred_block)) {
          blocks_to_visit.push(pred_block);
          blocks_visited.insert(pred_block);
        }
      }
    }
  }
}

bool Relaxer::runOnFunction(Function &F) {

  dom_tree = &getAnalysis<DominatorTree>();

  inst_id_map.clear();
  unsigned inst_counter = 0;
  for (Function::iterator BB = F.begin(), BBE = F.end(); BB != BBE; ++BB) {
    for (BasicBlock::iterator I = BB->begin(), IE = BB->end(); I != IE; ++I) {
      inst_id_map[I] = inst_counter;
      ++inst_counter;
    }
  }

  inst_vector.clear();
  inst_vector.resize(inst_counter);

  for (Function::iterator BB = F.begin(), BBE = F.end(); BB != BBE; ++BB) {
    for (BasicBlock::iterator I = BB->begin(), IE = BB->end(); I != IE; ++I) {

      CallInst *C = dyn_cast<CallInst>(I);
      if (C) {
        inst_vector.set(inst_id_map[I]);
        continue;
      }

      LoadInst *L = dyn_cast<LoadInst>(I);
      if (!L) continue;
      Value    *LPtr = L->getPointerOperand();

      for (Value::use_iterator UI = LPtr->use_begin(), UE = LPtr->use_end();
            UI != UE; ++UI) {

        StoreInst *S = dyn_cast<StoreInst>(*UI);
        if (!S) continue;
        Value     *SPtr = S->getPointerOperand();

        if (LPtr == SPtr && dom_tree->dominates(L, S))
          MarkInterveningInsts(L, S);
      }
    }
  }

  return false;
}
    
void Relaxer::print(std::ostream &O, const Module *M) const {

  for (DenseMap<Instruction *, unsigned>::const_iterator II =
       inst_id_map.begin(), IE = inst_id_map.end(); II != IE; ++II) {

    Instruction *I = II->first;
    bool         R = inst_vector.test(II->second);
    O << getStringId(I) << " " << R << "\n";
  }
}


