
#include "llvm/BasicBlock.h"
#include "llvm/Constants.h"
#include "llvm/Function.h"
#include "llvm/Instruction.h"
#include "llvm/Instructions.h"
#include "llvm/Module.h"
#include "llvm/Pass.h"
#include "llvm/Value.h"
#include "llvm/ADT/BitVector.h"
#include "llvm/Analysis/Dominators.h"
#include "llvm/Support/Debug.h"

#include "../FaultInjection/FaultInjectionUtils.h"

#include <stack>
#include <set>

using namespace vertical;
using namespace llvm;

class Relaxer2 : public FunctionPass {
private:
  bool runOnFunction(Function &F);
  bool doFinalization(Module &M);
  void getAnalysisUsage(AnalysisUsage &AU) const;

  DominatorTree *dom_tree;

public:
  static char ID;
  Relaxer2() : FunctionPass((intptr_t)&ID) {}
};


char Relaxer2::ID = 0;
static RegisterPass<Relaxer2> X("relaxer2", "Relaxer2");

void Relaxer2::getAnalysisUsage(AnalysisUsage &AU) const {
  AU.addRequired<DominatorTree>();
  AU.setPreservesAll();
}

static void InsertTransition(Constant *F, Instruction *I, 
                             int static_counter, Value *counter_ptr) {

  Value *dynamic_counter = new LoadInst(counter_ptr, "dynamic_counter", I);
  Value *counter = BinaryOperator::CreateAdd(
      dynamic_counter, ConstantInt::get(Type::Int32Ty, static_counter),
      "counter", I);
  CallInst::Create(F, counter, "", I);
  new StoreInst(ConstantInt::get(Type::Int32Ty, 0), counter_ptr, I);
}

bool Relaxer2::runOnFunction(Function &F) {

  dom_tree = &getAnalysis<DominatorTree>();

  // allocate the instruction counter (will be register-allocated by mem2reg)
  BasicBlock *entry = F.begin();
  BasicBlock::iterator insert_pos = entry->begin();
  while (isa<AllocaInst>(insert_pos)) ++insert_pos;

  Value *counter_ptr = new AllocaInst(Type::Int32Ty, "counter_ptr", insert_pos);
  new StoreInst(ConstantInt::getNullValue(Type::Int32Ty), counter_ptr,
                insert_pos);

  Constant *relax_trans = F.getParent()->getOrInsertFunction(
      "__POKE__RelaxTrans", Type::VoidTy, Type::Int32Ty, NULL);

  for (Function::iterator BB = F.begin(), BBE = F.end(); BB != BBE; ++BB) {
    int static_counter = 0;

    BasicBlock::iterator I_begin = BB->begin();
    if (BB == F.begin()) {
      I_begin++; /* alloca and store */
      I_begin++;
    }

    for (BasicBlock::iterator I = I_begin, IE = BB->end(); I != IE; ++I) {

      /* even though we insert the call before the side-effecting instruction,
       * it's as if we put it after */
      static_counter++;

      CallInst   *C = dyn_cast<CallInst>(I);
      ReturnInst *R = dyn_cast<ReturnInst>(I);
      if (C || R) {
        InsertTransition(relax_trans, I, static_counter, counter_ptr);
        static_counter = 0;
        continue;
      }

      StoreInst *S = dyn_cast<StoreInst>(I);
      if (!S) continue;
      Value     *SPtr = S->getPointerOperand();

      for (Value::use_iterator UI = SPtr->use_begin(), UE = SPtr->use_end();
            UI != UE; ++UI) {

        LoadInst *L = dyn_cast<LoadInst>(*UI);
        if (!L) continue;
        Value    *LPtr = L->getPointerOperand();

        if (LPtr == SPtr && dom_tree->dominates(L, S)) {
          InsertTransition(relax_trans, I, static_counter, counter_ptr);
          static_counter = 0;
          break;
        }
      }
    }

    if (static_counter > 0) {
      Instruction *InsertPos = BB->getTerminator();
      Value *OldCounter = new LoadInst(counter_ptr, "OldCounter", InsertPos);
      Value *NewCounter = BinaryOperator::CreateAdd(
          OldCounter, ConstantInt::get(Type::Int32Ty, static_counter),
          "NewCounter", InsertPos);
      new StoreInst(NewCounter, counter_ptr, InsertPos);
    }
  }

  return true;
}


bool Relaxer2::doFinalization(Module &M) {

  Function *MainFn = M.getFunction("main");
  BasicBlock::iterator I = MainFn->begin()->begin();

  // Skip over any allocas in main's entry block.
  while (isa<AllocaInst>(I)) ++I;

  // call the poke initialization call
  Constant *RelaxInitFn = M.getOrInsertFunction(
      "__POKE__RelaxInit", Type::VoidTy, NULL);
  CallInst::Create(RelaxInitFn, "", I);

  return true;
}
    
