
import optparse, shutil, os, sys, inspect

import poke
from poke import shell, log, compiler

log.init("setup.log")
shell.init(log, True)

class AppBuilder(object):

  def __init__(self, dir):

    self.parser = optparse.OptionParser(
        usage="usage: %prog [options]")

    self.parser.add_option("--clean",
        action="store_true", dest="clean",
        help="only clean generated files")

    self.parser.add_option("--noclean",
        action="store_true", dest="noclean",
        help="do not clean generated files")

    self.parser.add_option("--save-profile",
        action="store_true", dest="saveProfile",
        help="don't clean profile information")

    self.parser.add_option("--dbg",
        action="store_true", dest="dbg",
        help="build a debug build")

    self.parser.add_option("--gprof",
        action="store_true", dest="gprof",
        help="build a with gprof output")

    self.parser.add_option("--config",
        dest="config", default='test',
        help="application configuration")

    self.dir = dir
    self.name = os.path.basename(dir)

    self.bytecodeDir    = "%s/bytecode" % self.dir
    self.graphsDir      = "%s/graphs" % self.dir
    self.inputsDir      = "%s/inputs" % self.dir
    self.srcDir         = "%s/src" % self.dir
    self.evalDir        = "%s/eval" % self.dir
    self.gprofDir       = "%s/gprof" % self.dir

    self.configFile     = "%s/%s.config" % (
        self.dir,
        self.name)
    self.bytecodeFile   = "%s/%s.bc" % (
        self.bytecodeDir,
        self.name)
    self.binaryFile     = "%s/%s" % (
        self.bytecodeDir,
        self.name)
    self.gprofFile      = "%s/%s" % (
        self.gprofDir,
        self.name)
    self.evalFile       = "%s/%s.eval" % (
        self.evalDir,
        self.name)

    # defaults
    self.bytecodeInputs = ""
    self.evalInputs = ""
    self.hooks = 0

    # libraries used by poke
    self.gccLibFlags = "-L%s " % (
        os.popen('poke-config --llvm-libdir').readline().strip())
    self.gccLibFlags += "-lpoke_runtime "
    self.gccLibFlags += "%s/profile_rt.a " % (
      os.popen('llvm-config --libdir').readline().strip())



  def processArgs(self):

    self.options, self.args = self.parser.parse_args()
    self.profileString = "%s-%s" % (self.name, self.options.config)
    self.profileDir    = "%s/profile/%s" % (self.dir, self.profileString)

  def setup(self):

    self.processArgs()

    if not self.options.noclean:
      self.clean()

      if self.options.clean:
        sys.exit(0)

    if self.options.gprof:
      self.buildGProf()
      sys.exit(0)

    self.buildInputs()
    self.buildSrc()
    self.buildBytecode()
    self.buildGraphs()
    self.buildBinary()
    self.buildConfig()


  def clean(self):

    dirs = [self.configFile, self.bytecodeDir, self.gprofDir, self.graphsDir]

    if not self.options.saveProfile:
      pass
      #dirs.append(self.profileDir)

    for item in dirs:
      log.displayLine("removing %s..." % item)
      shell.check_call("rm -rf %s" % item)

  def buildInputs(self):
    pass

  def buildSrc(self):
    pass

  def buildBytecode(self):
    shell.check_call("mkdir -p %s" % self.bytecodeDir)
    pass

  def buildGProf(self):
    shell.check_call("mkdir -p %s" % self.gprofDir)
    pass

  def buildGraphs(self):

    log.displayLine("generating graphs...")
    shell.check_call("mkdir -p %s" % self.graphsDir)

    poke.pushd(self.graphsDir)
    shell.check_call("opt -dot-callgraph %s"  % self.bytecodeFile,
        stdout=shell.VOID, stderr=shell.STDOUT)
    shell.check_call("opt -dot-cfg-only %s"  % self.bytecodeFile,
        stdout=shell.VOID, stderr=shell.STDOUT)
    poke.popd()

  def buildConfig(self):

    log.displayLine("generating %s..." % self.configFile)

    configFileObj = open(self.configFile, 'w')
    configFileObj.write("""
# The path to the bytecode executable
BytecodeFile: %(bytecodeFile)s

# Inputs to copy to the executable directory (if any).
BytecodeInputs: %(bytecodeInputs)s

# Path to the evaluation program
EvalFile: %(evalFile)s

# Inputs to copy to the evaluation program directory (if any).
EvalInputs: %(evalInputs)s

# OS-native libraries required to generate a native binary of the bytecode file (if any).
GccLibFlags: %(gccLibFlags)s

# Set to '1' if the program calls poke_instrumentation_{start|end}.
# Otherwise, set to '0' (profiling will be always-on).
ProfilingHooks: %(hooks)s

# Directory prefix for condor execution and output
ProfileString: %(profileString)s

""" % dict(inspect.getmembers(self)))

    configFileObj.close()

  def buildBinary(self):

    log.displayLine("generating binary executable...")

    compiler.compile(self.bytecodeFile, self.binaryFile,
        self.gccLibFlags.split())

