#!/usr/bin/env python

import os, sys, optparse, re, shutil, tarfile
from string import Template

import poke, apps
from poke import configuration, shell, log
from apps.LineModifier import LineModifier

# TODO, make this a switch
rootDir = os.environ['PARSEC_ROOT']

os.environ["PARSECDIR"]  = rootDir
os.environ["PARSECPLAT"] ='i686-Linux.poke'

class Component:

  def __init__(self, dir):

    self.dir = dir
    self.name = os.path.basename(dir)

    self.srcDir = "%s/src" % os.getcwd()
    self.objDir = "%s/obj/%s" % (dir, os.environ["PARSECPLAT"])
    self.instDir = "%s/inst/%s" % (dir, os.environ["PARSECPLAT"])

  def build(self, options, gprof):

    self.options = options

    if gprof:
      os.environ.update({
        "CPP":      "cpp",
        "AR":       "ar",
        "RANLIB":   "ranlib",
        "CC":       "gcc",
        "CFLAGS":   "-O2 -g -pg -funroll-loops",
        "CXX":      "g++",
        "CXXFLAGS": "-O2 -g -pg -funroll-loops -fpermissive -fno-exceptions"})
    else:
      os.environ.update({
        "CPP":      "cpp",
        "AR":       "llvm-ar",
        "RANLIB":   "llvm-ranlib",
        "CC":       "llvm-gcc",
        "CFLAGS":   "-O2 -funroll-loops",
        "CXX":      "llvm-g++",
        "CXXFLAGS": "-O2 -funroll-loops -fpermissive -fno-exceptions"})

    if options.dbg:
      os.environ["CFLAGS"]   += " -O0 -g"
      os.environ["CXXFLAGS"] += " -O0 -g"

    self.configure(gprof)

    self.make()

  def configure(self, gprof):

    # parse the PARSEC bldconf file
    bldConfig = configuration.parseFile("%s/parsec/gcc.bldconf" % self.dir,
        build_conf=None,
        build_env=None,
        build_inplace=None)

    bldConfig.eval()

    # read build configuration variables and interpolate existing variables
    bldConfig.build_conf = Template(bldConfig.build_conf
        ).substitute(poke.BlankMap(os.environ, bldConfig.__dict__))

    regex=r'(\w+)=(?:"(.*?)"|(\S+))'
    for match in re.finditer(regex, bldConfig.build_env):
      os.environ[match.group(1)] = Template(
          match.group(match.lastindex)
          ).substitute(poke.BlankMap(os.environ))

  
    # build in place or build using separate object directory?
    log.displayLine("creating %s..." % self.objDir)
    os.system("mkdir -p %s" % self.objDir)

    if bldConfig.build_inplace == "TRUE":
      buildSrcDir = self.objDir
      shutil.rmtree(self.objDir)
      shutil.copytree(self.srcDir, self.objDir)
    else:
      assert bldConfig.build_inplace == "FALSE"
      buildSrcDir = self.srcDir

    # configure
    if os.path.exists("%s/configure" % buildSrcDir):

      poke.pushd(self.objDir)

      log.displayLine("running configure...")
      shell.check_call("%s/configure %s --prefix=%s" % (
        buildSrcDir, bldConfig.build_conf, self.instDir))

      poke.popd()

  def make(self):

    log.displayLine("running make...")

    poke.pushd(self.objDir)
    shell.check_call("make")
    poke.popd()


class Application(Component):

  def __init__(self, dir, binaryFile):
    Component.__init__(self, dir)
    self.inputsDir = "%s/inputs" % dir
    self.bytecodeFile = "%s/%s.bc" % (self.objDir, binaryFile)
    self.gprofFile = "%s/%s" % (self.objDir, binaryFile)


class Dependency(Component):

  def build(self, options):
    Component.build(self, options)
    self.makeInstall()

  def makeInstall(self):
    log.displayLine("running make install...")
    shell.check_call("make -C %s install" % self.objDir)



class AppBuilder(apps.AppBuilder):

  def __init__(self, dir, parsecApp, *parsecDeps):

    apps.AppBuilder.__init__(self, dir)

    self.parsecApp = parsecApp
    self.parsecComponents = parsecDeps + (parsecApp,)
    self.hooks = 1
    self.gccLibFlags = "-lpthread -lm -lstdc++"

  def processArgs(self):

    apps.AppBuilder.processArgs(self)

  def clean(self):

    apps.AppBuilder.clean(self)

    log.displayLine("removing symbolic links...")

#        if os.path.lexists(self.inputsDir):
#            os.remove(self.inputsDir)
#
#        if os.path.lexists(self.srcDir):
#            if os.path.islink(self.srcDir):
#                os.remove(self.srcDir)
#            else:
#                shutil.rmtree(self.srcDir)

    for component in self.parsecComponents:
      if os.path.exists(component.objDir):
        log.displayLine("removing build directory %s..." % component.objDir)
        shutil.rmtree(component.objDir)

#    def buildInputs(self):
#
#        if os.path.lexists(self.parsecApp.inputsDir):
#
#            # link inputs dir
#            # ---------------
#
#            log.displayLine("linking %s..." % self.inputsDir)
#
#            if os.path.lexists(self.inputsDir):
#                os.remove(self.inputsDir)
#            os.symlink(self.parsecApp.inputsDir, self.inputsDir)
#
#            inputPath = "%s/input_%s.tar" % (self.inputsDir,
#            self.options.quality)
#
#            if os.path.lexists(inputPath):
#
#                # unpack input file
#                # -----------------
#
#                log.displayLine("unpacking %s..." % inputPath)
#
#                tarFile = tarfile.open(inputPath)
#                for member in tarFile.getmembers():
#                    tarFile.extract(member, self.inputsDir)
#                tarFile.close()
#
#    def buildSrc(self):
#
#        # link src dir
#        # ------------
#
#        if len(self.parsecComponents) == 1:
#            log.displayLine("linking %s..." % self.srcDir)
#
#            if os.path.lexists(self.srcDir):
#                os.remove(self.srcDir)
#            os.symlink(self.parsecApp.srcDir, self.srcDir)
#        else:
#            log.displayLine("making %s..." % self.srcDir)
#            if not os.path.exists(self.srcDir):
#                os.mkdir(self.srcDir)
#
#            for component in self.parsecComponents:
#                srcDir = "%s/%s" % (self.srcDir, component.name)
#                log.displayLine("linking %s..." % srcDir)
#
#                if os.path.lexists(srcDir):
#                    os.remove(srcDir)
#                os.symlink(component.srcDir, srcDir)

  def buildBytecode(self):

    apps.AppBuilder.buildBytecode(self)

    for component in self.parsecComponents:
      component.build(self.options, False)

    # create symbolic link
    # --------------------

    log.displayLine("linking bytecode file %s..." % self.bytecodeFile)
    if os.path.lexists(self.bytecodeFile):
      os.remove(self.bytecodeFile)

    os.symlink(self.parsecApp.bytecodeFile, self.bytecodeFile)


  def buildGProf(self):

    apps.AppBuilder.buildGProf(self)

    for component in self.parsecComponents:
      component.build(self.options, True)

    # create symbolic link
    # --------------------

    log.displayLine("linking gprof file %s..." % self.gprofFile)
    if os.path.lexists(self.gprofFile):
      os.remove(self.gprofFile)

    os.symlink(self.parsecApp.gprofFile, self.gprofFile)


class Imagick(Dependency):

  dir = "%s/pkgs/libs/imagick" % rootDir

  def __init__(self):

    Dependency.__init__(self, self.dir)

  def configure(self):

    Dependency.configure(self)

    log.displayLine("transforming Makefiles...")

    lineModifiers = [
        LineModifier("^CFLAGS =",
          replacements={" -g ": " "},
          extension=" -emit-llvm"),
        LineModifier("^CCLD =",
          replacement="CCLD = llvm-ld -disable-opt"),
        LineModifier("--mode=link",
          replacements={
            '$(AM_CFLAGS)': "",
            '$(CFLAGS)': ""}),
        LineModifier("^CXXLINK =",
          replacements={
            '$(AM_CXXFLAGS)': "",
            '$(CXXFLAGS)': ""}, searchUpto = 2),
        LineModifier("^LINK =",
          replacements={
            '$(AM_CFLAGS)': "",
            '$(CFLAGS)': ""}, searchUpto = 2)]

    for root, dirs, files in os.walk(self.objDir):
      for file in files:
        if file == "Makefile":
          file = "%s/Makefile" % root
          log.displayLine("  %s" % file)
          LineModifier.replaceFile(lineModifiers, file)


class Gsl(Dependency):

  dir = "%s/pkgs/libs/gsl" % rootDir

  def __init__(self):

    Dependency.__init__(self, self.dir)

  def configure(self):

    Dependency.configure(self)

    log.displayLine("transforming Makefiles...")

    lineModifiers = [
        LineModifier("^CFLAGS =",
          extension=" -emit-llvm"),
        LineModifier("^CCLD =",
          replacement="CCLD = llvm-ld -disable-opt"),
        LineModifier("^LINK =",
          replacements={
            '$(AM_CFLAGS)': "",
            '$(CFLAGS)': ""}, searchUpto = 2)]

    for root, dirs, files in os.walk(self.objDir):
      for file in files:
        if file == "Makefile":
          file = "%s/Makefile" % root
          log.displayLine("  %s" % file)
          LineModifier.replaceFile(lineModifiers, file)


class Libxml2(Dependency):

  dir = "%s/pkgs/libs/libxml2" % rootDir

  def __init__(self):

    Dependency.__init__(self, self.dir)

  def configure(self):

    Dependency.configure(self)

    log.displayLine("transforming Makefiles...")

    lineModifiers = [
        LineModifier("^CFLAGS =",
          replacements={" -g ": " "},
          extension=" -emit-llvm"),
        LineModifier("^CCLD =",
          replacement="CCLD = llvm-ld -disable-opt"),
        LineModifier("^LINK =",
          replacements={
            '$(AM_CFLAGS)': "",
            '$(CFLAGS)': ""}, searchUpto = 2)]

    for root, dirs, files in os.walk(self.objDir):
      for file in files:
        if file == "Makefile":
          file = "%s/Makefile" % root
          log.displayLine("  %s" % file)
          LineModifier.replaceFile(lineModifiers, file)


class Glib(Dependency):

  dir = "%s/pkgs/libs/glib" % rootDir

  def __init__(self):

    Dependency.__init__(self, self.dir)

  def configure(self):

    Dependency.configure(self)

    log.displayLine("transforming Makefiles...")

    lineModifiers = [
        LineModifier("^CFLAGS =",
          replacements={" -g ": " "},
          extension=" -emit-llvm"),
        LineModifier("^CCLD =",
          replacement="CCLD = --tag=CXX llvm-ld -disable-opt"),
        LineModifier("^SUBDIRS =",
          replacements={
            "tests": "",
            "docs": ""}),
        LineModifier("^LINK =",
          replacements={
            '$(AM_CFLAGS)': "",
            '$(CFLAGS)': ""}, searchUpto = 2)]

    for root, dirs, files in os.walk(self.objDir):
      for file in files:
        if file == "Makefile":
          file = "%s/Makefile" % root
          log.displayLine("  %s" % file)
          LineModifier.replaceFile(lineModifiers, file)

      log.displayLine("transforming config.h...")

      lineModifiers = [LineModifier("#define G_ATOMIC", replacement="")]

      file = "%s/config.h" % self.objDir
      LineModifier.replaceFile(lineModifiers, file)

