
import os

import poke


class Function(object):

    def __init__(self, name):
        self.name = name
        self.blockList = []
        self.id = None

    def getScopedName(self):
        return self.name


class Block(object):

    def __init__(self, name, function):
        self.name = name
        self.function = function
        self.instList = []
        self.id = None

    def getScopedName(self):
        return "%s::%s" % (self.function.name, self.name)


class Instruction(object):

    def __init__(self, opcodeName, block, position):
        self.opcodeName  = opcodeName
        self.block       = block
        self.position    = position
        self.criticality = 0
        self.id          = None

    def getScopedName(self):
        return "%s::%s" % (self.block.getScopedName(), self.position)

    def getOpcode(self):
        # dynamically fetch opcode so that the opcode is not pickled
        return InstructionOpcode.getOpcode(self.opcodeName)

    def getType(self):
        return self.getOpcode().type

    def getTypeGroup(self):
        return self.getType().group


class InstructionTypeGroup(object):

    groupMap = {}

    def __init__(self, name):
        self.name           = name
        self.groupMap[name] = self
        self.typeList       = []

    def registerType(self, type):
        self.typeList.append(type)

    @staticmethod
    def getGroup(name):
        return InstructionTypeGroup.groupMap[name]

    @staticmethod
    def getGroups():
        return InstructionTypeGroup.groupMap.values()

    @staticmethod
    def getInstrumentedGroups(semantics):
        o = []
        for group in InstructionTypeGroup.getGroups():
            for type in group.typeList:
                if type.isInstrumented(semantics):
                    o.append(group)
                    break
        return o

    @staticmethod
    def getInstrumentedGroupCombinations(semantics):
        for groups in poke.powerSet(InstructionTypeGroup.getInstrumentedGroups(semantics)):
            groupNames = [group.name for group in groups]
            if not groupNames:
                continue

            yield ".".join(groupNames)

    
class InstructionType(object):

    typeMap = {}

    def __init__(self, name, instrumentedMap, groupName):
        self.name             = name
        self.instrumentedMap  = instrumentedMap
        self.opcodeList       = []
        self.group            = InstructionTypeGroup.getGroup(groupName)

        self.typeMap[name]  = self
        self.group.registerType(self)

    def registerOpcode(self, opcode):
        self.opcodeList.append(opcode)

    def isInstrumented(self, semantics):
        return self.instrumentedMap[semantics]

    @staticmethod
    def getType(name):
        return InstructionType.typeMap[name]

    @staticmethod
    def getTypes():
        return InstructionType.typeMap.values()

    @staticmethod
    def getInstrumentedTypes(semantics):
        o = []
        for type in InstructionType.getTypes():
            if type.instrumentedMap[semantics]:
                o.append(type)
        return o


class InstructionOpcode(object):

    opcodeMap = {}

    def __init__(self, name, typeName):
        self.name            = name
        self.type            = InstructionType.getType(typeName)

        self.opcodeMap[name] = self
        self.type.registerOpcode(self)

    @staticmethod
    def getOpcode(name):
        return InstructionOpcode.opcodeMap[name]

    @staticmethod
    def getOpcodes():
        return InstructionOpcode.opcodeMap.values()


InstructionTypeGroup("Control")
InstructionTypeGroup("Functional")
InstructionTypeGroup("Memory")
InstructionTypeGroup("Other")

instrumentedMap = {"recovery": True, "bitflip": True}
InstructionType("Arithmetic",    instrumentedMap, "Functional")
InstructionType("CastFpInt",     instrumentedMap, "Functional")
InstructionType("GetElementPtr", instrumentedMap, "Memory")
InstructionType("Load",          instrumentedMap, "Memory")
InstructionType("Logic",         instrumentedMap, "Functional")

instrumentedMap = {"recovery": False, "bitflip": False}
InstructionType("AllocFree",     instrumentedMap, "Memory")
InstructionType("CastTruncExt",  instrumentedMap, "Functional")
InstructionType("CastTypeConv",  instrumentedMap, "Functional")
InstructionType("Function",      instrumentedMap, "Control")
InstructionType("PHI",           instrumentedMap, "Other")
InstructionType("UncondBranch",  instrumentedMap, "Control")
InstructionType("Unreachable",   instrumentedMap, "Other")
InstructionType("VAArgs",        instrumentedMap, "Other")
InstructionType("Vector",        instrumentedMap, "Other")

instrumentedMap = {"recovery": True, "bitflip": False}
InstructionType("CondBranch",    instrumentedMap, "Control")
InstructionType("Compare",       instrumentedMap, "Control")
InstructionType("Select",        instrumentedMap, "Control")
InstructionType("Store",         instrumentedMap, "Memory")

for opcode in ["Alloca", "Malloc", "Free"]:
    InstructionOpcode(opcode, "AllocFree")

for opcode in ["Add", "Sub", "Mul", "UDiv", "SDiv", "FDiv", "URem", "SRem", "FRem"]:
    InstructionOpcode(opcode, "Arithmetic")

for opcode in ["FPToUI", "FPToSI", "UIToFP", "SIToFP"]:
    InstructionOpcode(opcode, "CastFpInt")

for opcode in ["Trunc", "ZExt", "SExt", "FPTrunc", "FPExt"]:
    InstructionOpcode(opcode, "CastTruncExt")

for opcode in ["PtrToInt", "IntToPtr", "BitCast"]:
    InstructionOpcode(opcode, "CastTypeConv")

for opcode in ["ICmp", "FCmp"]:
    InstructionOpcode(opcode, "Compare")

for opcode in ["Br", "Switch"]:
    InstructionOpcode(opcode, "CondBranch")

for opcode in ["Invoke", "Call", "Ret", "Unwind"]:
    InstructionOpcode(opcode, "Function")

for opcode in ["GetElementPtr"]:
    InstructionOpcode(opcode, "GetElementPtr")

for opcode in ["Load"]:
    InstructionOpcode(opcode, "Load")

for opcode in ["Shl", "LShr", "AShr", "And", "Or", "Xor"]:
    InstructionOpcode(opcode, "Logic")

for opcode in ["PHI"]:
    InstructionOpcode(opcode, "PHI")

for opcode in ["Select"]:
    InstructionOpcode(opcode, "Select")

for opcode in ["Store"]:
    InstructionOpcode(opcode, "Store")

for opcode in ["UncondBr"]:
    InstructionOpcode(opcode, "UncondBranch")

for opcode in ["Unreachable"]:
    InstructionOpcode(opcode, "Unreachable")

for opcode in ["VAArgs"]:
    InstructionOpcode(opcode, "VAArgs")

for opcode in ["ExtractElement", "InsertElement", "ShuffleVector"]:
    InstructionOpcode(opcode, "Vector")





