
import os, sys, re, operator, signal

import poke
from poke import log, shell


# globals
tagsFiles = []
srcDir = ""


def getLang(context):
  if "-lstdc++" in context.gccLibFlags:
    return "C++"
  else:
    return "C"

def buildTags(context):
  global tagsFiles, srcDir

  srcDir = os.path.normpath("%s/../src" % os.path.dirname(context.bytecodeFile))

  appTagsFile = "%s/tags" % srcDir
  tagsFiles.append(appTagsFile)

  if getLang(context) == "C++":
    cppTagsFile = "%s/../include/c++/%s/tags" % (
        os.path.dirname(os.popen("which llvm-gcc").readline().strip()),
        os.popen("llvm-gcc --version").readline().split()[2])

    tagsFiles.append(cppTagsFile)

  for tagsFile in tagsFiles:
    if not os.path.exists(tagsFile):
      poke.pushd()
      os.chdir(os.path.dirname(tagsFile))

      print(" building tags file %s..." % tagsFile)
      shell.check_call(
          "ctags -R --languages=C,C++ --extra=+fq --fields=+inS -f %s" %
          tagsFile)

      poke.popd()

  extraTagsFile = "%s/tags.poke" % srcDir
  if os.path.exists(extraTagsFile):
      tagsFiles.insert(0, extraTagsFile)


def stripTemplates(str):
    nesting = 0
    chars = []
    for char in str:
      if char == '<':    nesting += 1
      elif char == '>':  nesting -= 1
      elif nesting == 0: chars.append(char)
    return "".join(chars)


def getCallGraph(file, func):

  class Node:
    def __init__(self):
      self.callers = []
      self.name = ""
      self.keep = False

  nodeMap = {}

  nodeNameRegex = re.compile('\s+(Node\S+).*label="\{([^|\}]+)')
  nodeArrowRegex = re.compile('\s+(Node[^:]+).*(Node[^;]+)')

  keepList = []

  # read in the file
  # ----------------

  inFileObj = open(file)
  for line in inFileObj:

    match = nodeNameRegex.match(line)
    if match:
      node = nodeMap.setdefault(match.group(1), Node())
      node.name = match.group(2)
      if node.name == func:
        node.keep = True
        keepList.append(node)
      continue
        
    match = nodeArrowRegex.match(line)
    if match:
      caller = nodeMap.setdefault(match.group(1), Node())
      callee = nodeMap.setdefault(match.group(2), Node())
      callee.callers.append(caller)

  inFileObj.close()

  # what do we want to keep?
  # ------------------------

  index = 0
  while index < len(keepList):
    node = keepList[index]
    keepList.extend(node.callers)
    for caller in node.callers:
      if caller.name != "Indirect call node":
        caller.keep = True
    index += 1

  # make it happen
  # --------------

  outFile = poke.tempFile()

  outFileObj = open(outFile, "w")
  inFileObj = open(file)
  for line in inFileObj:

    match = nodeNameRegex.match(line)
    if match:
      if not nodeMap[match.group(1)].keep:
        continue
      else:
        line = line.replace(match.group(2),
            stripTemplates(os.popen("c++filt %s" % match.group(2)).readline().strip()))

    match = nodeArrowRegex.match(line)
    if match:
      if (not nodeMap[match.group(1)].keep or
          not nodeMap[match.group(2)].keep):
        continue
    
    outFileObj.write(line)

  inFileObj.close()
  outFileObj.close()

  return outFile


def findTag(name, params):

  srcFile = None
  srcLine = None

  for tagsFile in tagsFiles:

    def truncateNum(str):
      return re.sub("^(.*?)\d+$", r"\1", str)

    def truncateClass(str):
      return re.sub("^.*?::(.*?)$", r"\1", str)

    searchNames = set([
      name,
      truncateNum(name),
      truncateClass(name),
      truncateClass(truncateNum(name))])

    for searchName in searchNames:

      log.write("searching for tag %s in file %s\n" % (
        searchName, tagsFile))

      tagRegex = re.compile(
          '^%s\s+(?P<file>\S+)\s+.*f\s+line:(?P<line>\d+).*signature:\((?P<sig>.*)\)$' %
          searchName)

      tagsFileObj = open(tagsFile)
      for index, line in enumerate(tagsFileObj):

        match = tagRegex.match(line)
        if match:
          log.write("function %s matches on line %s\n" % (searchName, index + 1))

          signature = stripTemplates(match.group(3))
          if (params == None or
              len(signature.split(",")) == len(params.split(","))):

            srcFile = "%s/%s" % (os.path.dirname(tagsFile), match.group(1))
            srcLine = match.group(2)
            print
            print(" in file %s:%s:" % (srcFile, srcLine))
            break

      tagsFileObj.close()

      if srcFile:
        break

    return srcFile, srcLine


def locate(instIds, context):

  log.writeHeader("Locating instructions")

  functionMap = {}

  for instId in instIds:
    inst = context.stats.instructionList[int(instId)]
    functionMap.setdefault(inst.block.function, []).append(instId)

  for function in sorted(functionMap, key=operator.attrgetter("id")):
      
    instIdList = functionMap[function]

    demangledFunction = stripTemplates(
            os.popen("c++filt %s" %
              function.name).readline().strip())

    match = re.match(
        "(?:(?P<ret>[^\(\s]+)\s)?(?P<name>[^\(]+)(?:\((?P<params>[^\)]*)\))?",
        demangledFunction)

    ret, name, params = match.groups()

    log.write("ret: %s\n" % ret)
    log.write("name: %s\n" % name)
    log.write("params: %s\n" % params)
    log.write("\n")

    srcFile, srcLine = findTag(name, params)

    assert srcFile, "tag for %s (%s) not found!" % (name, params)

    for instId in instIdList:
      inst = context.stats.instructionList[int(instId)]
      print("  (%s, %s) %s" % (instId, inst.opcodeName, inst.getScopedName()))

    if context.inspect:

      # find the right line in the .ll file
      defRegex = re.compile('^define.*%s' % function.name)

      llFile = context.bytecodeFile.replace(".bc", ".ll")
      llLine = None
      llFileObj = open(llFile)
      for index, line in enumerate(llFileObj):

          match = defRegex.match(line)
          if match:
              llLine = index + 1
              break

      assert llLine, ".ll line number not found!"
      llFileObj.close()

      poke.pushd(srcDir)

      try:
        src = shell.Popen(
            "gvim -f -R +%s %s -geom 100x75 -c 'winpos 10 10' \
                --cmd 'set tags+=%s'" % (
                  srcLine, srcFile, ",".join(tagsFiles)))
      finally:
        poke.popd()

      ll = shell.Popen(
          "gvim -f -R +%s %s -geom 100x75 -c 'winpos 640 10'" % (
            llLine, llFile))

      graphsDir = os.path.dirname(context.bytecodeFile) + "/../graphs"

      callGraphFile = getCallGraph("%s/callgraph.dot" % graphsDir, function.name)
      psFile1 = poke.tempFile()
      psFile2 = poke.tempFile()

      try:
        shell.check_call("dot %s -Tps > %s" % (
          callGraphFile, psFile1))
        ps1 = shell.Popen("psview -magstep -3 -killkids %s" % psFile1)

        shell.check_call("dot %s/cfg.%s.dot -Tps > %s" % (
          graphsDir, function.name, psFile2))
        ps2 = shell.Popen("psview -magstep -3 -killkids %s" % psFile2)

        print 
        print(" Continue? [Y/N]"),

        input = ""
        while input != "y" and input != "n":
          input = raw_input().strip().lower()

        ps1.kill(signal.SIGHUP)
        ps2.kill(signal.SIGHUP)

        # kill lingering "gs" processes invoked by psview
        shell.check_call("ps | egrep ' gs$' | awk '{print $1}' | xargs kill -9")

        if input == "n":
          # finally still executed
          break 

      finally:
        ll.kill()
        src.kill()





