
import os, sys, shutil, re

import poke, compiler, log, shell, statistics, filewrap
from llvm import *


profilingLibPath = "%s/Profiling.so" % (
    os.popen('poke-config --llvm-libdir').readline().strip()) 

analysisLibPath = "%s/FaultInjection.so" % (
    os.popen('poke-config --llvm-libdir').readline().strip())

relaxerLibPath = "%s/Relaxer.so" % (
    os.popen('poke-config --llvm-libdir').readline().strip())

def llvm(context, quality, threads):

  stats = context.stats

  log.displayLine(
      "   generating LLVM dynamic profile for quality %s threads %s..." % (
        quality, threads))
  sys.stdout.flush()

  if context.debug:
    debugFlag = "--debug"
  else:
    debugFlag = ""

  # instrument with profiling hooks
  # -------------------------------

  runDir = "%s/profile/llvm" % os.getcwd()
  shell.check_call("mkdir -p %s" % runDir)
  shell.check_call("rm -rf %s/*" % runDir)

  instrumentedFile = poke.tempFile()

  optCommand = 'opt %s -f -load %s -block-profiling -count-llvm %s -o %s' % (
      debugFlag, profilingLibPath, context.bytecodeFile, instrumentedFile)

  if context.profilingHooks:
    optCommand += " -profiling-hooks"

  shell.check_call(optCommand)

  # compile 
  # -------

  poke.pushd(runDir)
  nativePath = os.path.basename(context.bytecodeFile).replace(".bc", "")
  compiler.compile(instrumentedFile, nativePath, context.gccLibFlags)
  poke.popd()

  # copy data
  # ---------

  poke.copyPaths(context.bytecodeInputs, runDir)

  # run the program
  # ---------------

  command = './%s %s %s' % (nativePath, quality, 2**threads)

  poke.pushd(runDir)
  if context.remoteHost:
    shell.check_call("ssh %s 'rm -rf ~/.scratch && mkdir ~/.scratch'" % (
      context.remoteHost))
    shell.check_call("scp -r * %s:~/.scratch" % (
      context.remoteHost))
    shell.check_call("ssh %s 'cd ~/.scratch && %s && cd -'" % (
      context.remoteHost, command))
    shell.check_call("scp %s:~/.scratch/llvmprof.out ." % (
      context.remoteHost))
  else:
    # change to temp directory so llvmprof.out is written
    shell.check_call(command)
  poke.popd()

  # read in the dynamic profiling information
  # -----------------------------------------

  profile = stats.profiles.setdefault(
      (quality, threads),
      statistics.Profile(stats))

  p = shell.Popen('%s/poke-profiler %s %s' % (
    os.popen('poke-config --bindir').readline().strip(),
    context.bytecodeFile, "%s/llvmprof.out" % runDir),
    stdout=shell.PIPE)

  p.stdout = filewrap.FileWrapper(p.stdout)

  p.stdout.advanceTo(re.compile("^Function execution frequencies:"))
  p.stdout.next()
  p.stdout.next()

  for line in p.stdout:
    info = line.split()

    if not info:
      break
    else:
      assert len(info) == 3

    functionExecs = int(info[1].split("/")[0])
    functionName  = info[2]
    functionId    = stats.getFunction(functionName).id  
    profile.functionExecs[functionId] = functionExecs

  p.stdout.advanceTo(re.compile("^Basic block execution frequencies:"))
  p.stdout.next()
  p.stdout.next()

  for line in p.stdout:
    info = line.split()
    assert len(info) == 6

    blockName    = info[5]
    blockExecs   = int(info[2].split("/")[0])
    functionName = info[3][:-2]
    blockId      = stats.getBlock(functionName, blockName).id
    profile.blockExecs[blockId] = blockExecs

  p.check_wait()

  # compute statistics
  # ------------------
  
  profile.totalFunctionExecs = sum(profile.functionExecs)
  profile.totalBlockExecs    = sum(profile.blockExecs)

  for instruction in stats.instructionList:
    profile.totalInstExecs += profile.blockExecs[instruction.block.id]
    profile.instExecsByTypeMap[instruction.getType().name] += (
        profile.blockExecs[instruction.block.id])



def pprof(context, quality, threads):

  stats = context.stats

  runDir = "%s/profile/pprof" % os.getcwd()
  shell.check_call("mkdir -p %s" % runDir)
  shell.check_call("rm -rf %s/*" % runDir)

  log.displayLine(
      "   generating Google dynamic profile for quality %s threads %s..." % (
        quality, threads))
  sys.stdout.flush()

  # compile 
  # -------

  if context.debug:
    debugFlag = "--debug"
  else:
    debugFlag = ""

  if not context.profilingHooks:
    # this will insert the hooks when the program starts and exits
    instrumentedFile = poke.tempFile()
    shell.check_call('opt %s -f -load %s -block-profiling %s -o %s' % (
      debugFlag, profilingLibPath, context.bytecodeFile, instrumentedFile))
  else:
    instrumentedFile = context.bytecodeFile

  poke.pushd(runDir)
  nativePath = os.path.basename(context.bytecodeFile).replace(".bc", "")

  compiler.compile(instrumentedFile, nativePath, context.gccLibFlags, True)
  poke.popd()

  # copy data
  # ---------

  poke.copyPaths(context.bytecodeInputs, runDir)

  # run the program
  # ---------------

  command = 'PPROF=./%s.prof ./%s %s %s' % (
      nativePath, nativePath, quality, 2**threads)

  poke.pushd(runDir)
  if context.remoteHost:
    shell.check_call("ssh %s 'rm -rf ~/.scratch && mkdir ~/.scratch'" % (
      context.remoteHost))
    shell.check_call("scp -r * %s:~/.scratch" % (
      context.remoteHost))
    shell.check_call("ssh %s 'cd ~/.scratch && %s && cd -'" % (
      context.remoteHost, command),
      stderr=shell.PIPE)
  else:
    shell.check_call(command, stderr=shell.PIPE)
  poke.popd()


def relaxer(context, quality, threads):

  runDir = "%s/profile/relaxer" % os.getcwd()
  shell.check_call("mkdir -p %s" % runDir)
  shell.check_call("rm -rf %s/*" % runDir)

  log.displayLine(
      "   generating Relaxer dynamic profile for quality %s threads %s..." % (
        quality, threads))
  sys.stdout.flush()

  # compile 
  # -------

  if context.debug:
    debugFlag = "--debug"
  else:
    debugFlag = ""

  # this will insert the hooks when the program starts and exits
  instrumentedFile = poke.tempFile()
  shell.check_call('opt %s -f -load %s -relaxer2 %s -o %s' % (
    debugFlag, relaxerLibPath, context.bytecodeFile, instrumentedFile))

  poke.pushd(runDir)
  nativePath = os.path.basename(context.bytecodeFile).replace(".bc", "")

  compiler.compile(instrumentedFile, nativePath, context.gccLibFlags, True)
  poke.popd()

  # copy data
  # ---------

  poke.copyPaths(context.bytecodeInputs, runDir)

  # run the program
  # ---------------

  poke.pushd(runDir)
  if context.remoteHost:
    shell.check_call("ssh %s 'rm -rf ~/.scratch && mkdir ~/.scratch'" % (
      context.remoteHost))
    shell.check_call("scp -r * %s:~/.scratch" % (
      context.remoteHost))
  poke.popd()

  command = './%s %s %s' % (
      nativePath, quality, 2**threads)

  poke.pushd(runDir)
  if context.remoteHost:
    p = shell.Popen("ssh %s 'cd ~/.scratch && %s && cd -'" % (
      context.remoteHost, command),
      stderr=shell.PIPE)
  else:
    p = shell.Popen(command, stderr=shell.PIPE)
  poke.popd()

  p.check_wait()

  if context.remoteHost:
    shell.check_call("ssh %s 'rm -rf ~/.scratch'" % (
      context.remoteHost))


def timing(context, quality, threads):

  stats = context.stats
  profile = stats.profiles.setdefault(
      (quality, threads),
      statistics.Profile(stats))

  runDir = "%s/profile/timing" % os.getcwd()
  shell.check_call("mkdir -p %s" % runDir)
  shell.check_call("rm -rf %s/*" % runDir)

  log.displayLine(
      "   generating Timing dynamic profile for quality %s threads %s..." % (
        quality, threads))
  sys.stdout.flush()

  # compile 
  # -------

  if context.debug:
    debugFlag = "--debug"
  else:
    debugFlag = ""

  if not context.profilingHooks:
    # this will insert the hooks when the program starts and exits
    instrumentedFile = poke.tempFile()
    shell.check_call('opt %s -f -load %s -block-profiling %s -o %s' % (
      debugFlag, profilingLibPath, context.bytecodeFile, instrumentedFile))
  else:
    instrumentedFile = context.bytecodeFile

  poke.pushd(runDir)
  nativePath = os.path.basename(context.bytecodeFile).replace(".bc", "")

  compiler.compile(instrumentedFile, nativePath, context.gccLibFlags, True)
  poke.popd()

  # copy data
  # ---------

  poke.copyPaths(context.bytecodeInputs, runDir)

  # run the program
  # ---------------

  poke.pushd(runDir)
  if context.remoteHost:
    shell.check_call("ssh %s 'rm -rf ~/.scratch && mkdir ~/.scratch'" % (
      context.remoteHost))
    shell.check_call("scp -r * %s:~/.scratch" % (
      context.remoteHost))
  poke.popd()

  for i in range(5):
    command = './%s %s %s' % (
        nativePath, quality, 2**threads)

    poke.pushd(runDir)
    if context.remoteHost:
      p = shell.Popen("ssh %s 'cd ~/.scratch && %s && cd -'" % (
        context.remoteHost, command),
        stderr=shell.PIPE)
    else:
      p = shell.Popen(command, stderr=shell.PIPE)
    poke.popd()

    time = ""
    p.stderr = log.wrap(p.stderr)
    for line in p.stderr:
      if line.startswith("__POKE__ Time:"):
        time = line.split()[2]

    assert(time)
    profile.times.append(float(time))

    p.check_wait()

  if context.remoteHost:
    shell.check_call("ssh %s 'rm -rf ~/.scratch'" % (
      context.remoteHost))


def staticAnalyze(context):

  stats = context.stats
  stats.clearStaticData()

  log.displayLine("   generating static profile...")
  sys.stdout.flush()

  if context.debug:
    debugFlag = "--debug"
  else:
    debugFlag = ""

  # compile static function/block information from bytecode file
  # --------------------------------------------------------------

  p = shell.Popen('opt %s -load %s -real-translate -snapshot -analyze %s' % (
    debugFlag, analysisLibPath, context.realFile),
    stdout=shell.PIPE, stderr=shell.VOID)

  functionRegex    = re.compile("Function (\S+):")
  blockRegex       = re.compile(" BasicBlock (\S+):")
  instructionRegex = re.compile("  (\w+)")

  for line in p.stdout:
        
    match = instructionRegex.match(line)
    if match:
      opcode = match.group(1)
      if opcode == "Skip":
        skipped += 1
        continue

      positionInBlock = len(block.instList) + skipped + 1
      instruction     = Instruction(opcode, block, positionInBlock)

      block.instList.append(instruction)
      stats.registerInstruction(instruction)
      continue

    match = blockRegex.match(line)
    if match:
      blockName = match.group(1)
      block     = Block(blockName, function)
      skipped   = 0

      function.blockList.append(block)
      stats.registerBlock(block)
      continue

    match = functionRegex.match(line)
    if match:
      functionName = match.group(1)
      function     = Function(functionName)

      stats.registerFunction(function)
      continue

  p.check_wait()

  # translation analysis
  # --------------------

  p = shell.Popen('opt %s -load %s -real-translate -analyze %s' % (
    debugFlag, analysisLibPath, context.realFile),
    stdout=shell.PIPE, stderr=shell.VOID)

  p.stdout.next() 

  for inst in stats.instructionList:
    inst.instrumented = 0

  for block in stats.blockList:
    block.relax = 0

  for func in stats.functionList:
    func.instrumented = 0

  for line in p.stdout:
    if line.startswith("Printing analysis"):
      continue

    if line.startswith("Entry Block:"):
      blockName = line.split(": ")[1].strip()
      block = stats.getBlock(*blockName.split("::"))
      block.relax = 1
      continue

    inst = stats.getInstruction(*line.strip().split("::"))
    inst.instrumented = 1
    inst.block.function.instrumented = 1

  p.check_wait()

  # Relaxer analysis
  # ----------------

  p = shell.Popen('opt %s -load %s -relaxer -analyze %s' % (
    debugFlag, relaxerLibPath, context.bytecodeFile),
    stdout=shell.PIPE, stderr=shell.VOID)

  for line in p.stdout:
    if line.startswith("Printing analysis"):
      continue
    instName, relaxer = line.split()
    try:
      inst = stats.getInstruction(*instName.split("::"))
      inst.relaxer = int(relaxer)
    except KeyError:
      pass

  p.check_wait()


  return stats


