import os, sys, shutil, re, cPickle

import poke, log, shell, profiler
from llvm import *


class Profile(object):

  def __init__(self, stats):

    self.stats = stats
    self.times = []

    self.functionExecs = []
    for i in range(len(stats.functionList)):
      self.functionExecs.append(0)

    self.blockExecs = []
    for i in range(len(stats.blockList)):
      self.blockExecs.append(0)

    self.instructionExecs = []
    for i in range(len(stats.instructionList)):
      self.instructionExecs.append(0)

    self.totalFunctionExecs = 0
    self.totalBlockExecs = 0
    self.totalInstExecs = 0

    self.instExecsByTypeMap = {}

    for type in InstructionType.getTypes():
      self.instExecsByTypeMap[type.name] = 0

  def getInstExecs(self, what=None):
    if what:
      if isinstance(what, InstructionType):
        return self.instExecsByTypeMap[what.name]
      if isinstance(what, Function):
        return sum(self.blockExecs[b.id] * len(b.instList)
            for b in what.blockList)
    
    return self.totalInstExecs

  def writeSubsetsGraph(self, fileObj):

    staAccum = 0
    dynAccum = 0
    for fn in sorted(self.stats.functionList,
        key=lambda f: sum(self.blockExecs[b.id] * len (b.instList)
          for b in f.blockList),
        reverse=True):

      staAccum += sum(len(b.instList) for b in fn.blockList)
      dynAccum += sum(self.blockExecs[b.id] * len(b.instList)
          for b in fn.blockList)

      fileObj.write("%s %s\n" % (
        poke.fractionToPercent(dynAccum, self.totalInstExecs),
        staAccum))


class Stats(object):

  def __init__(self):
    self.profiles = {}
    self.clearStaticData()

  def clearStaticData(self):
    self.functionMap = {}
    self.blockMap = {}
    self.instructionMap = {}

    self.functionList = []
    self.blockList = []
    self.instructionList = []

    self.instructionCounter = 0
    self.blockCounter = 0
    self.functionCounter = 0

    self.instCountByTypeMap = {}
    for type in InstructionType.getTypes():
      self.instCountByTypeMap[type.name] = 0

  
  def registerInstruction(self, inst):
    self.instructionMap[inst.getScopedName()] = inst
    self.instructionList.append(inst)
    inst.id = self.instructionCounter
    self.instructionCounter += 1
    self.instCountByTypeMap[inst.getType().name] += 1

  def registerBlock(self, block):
    self.blockMap[block.getScopedName()] = block
    self.blockList.append(block)
    block.id = self.blockCounter
    self.blockCounter += 1

  def registerFunction(self, function):
    self.functionMap[function.getScopedName()] = function
    self.functionList.append(function)
    function.id = self.functionCounter
    self.functionCounter += 1

  def getInstruction(self, function, block, position):
    return self.instructionMap["%s::%s::%s" % (function, block, position)]

  def getBlock(self, function, block):
    return self.blockMap["%s::%s" % (function, block)]

  def getFunction(self, function):
    return self.functionMap["%s" % (function)]


  def save(self, saveDir):
    
    shell.check_call("mkdir -p %s" % saveDir)

    # pickle
    pickleFile = open("%s/stats.pickle" % saveDir, 'w')
    cPickle.dump(self, pickleFile, 2)
    pickleFile.close()

    # text dump
    for (quality, threads), profile in self.profiles.iteritems():
      profileDir = "%s/%s_%s" % (saveDir, quality, threads)
      shell.check_call("mkdir -p %s" % profileDir)

      # listing
      listingFile = open('%s/listing.txt' % profileDir, 'w')

      listingFile.write("\n")
      listingFile.write("%-34s %15s\n" % (
        "Static Function Count:", len(self.functionList)))
      listingFile.write("%-34s %15s\n" % (
        "Static Basic Block Count:", len(self.blockList)))
      listingFile.write("%-34s %15s\n" % (
        "Static Instruction Count:", len(self.instructionList)))

      listingFile.write("\n")
      listingFile.write("%-34s %15s\n" % (
        "Dynamic Function Executions:", profile.totalFunctionExecs))
      listingFile.write("%-34s %15s\n" % (
        "Dynamic Basic Block Executions:", profile.totalBlockExecs))
      listingFile.write("%-34s %15s\n" % (
        "Dynamic Instruction Executions:", profile.totalInstExecs))

      if profile.times:
        listingFile.write("\n")
        for trial, time in enumerate(profile.times):
          listingFile.write("%-34s %15s\n" % (
            "Execution Time Trial %s:" % trial, time))
        listingFile.write("%-34s %15s\n" % (
          "Min MIPS:", profile.totalInstExecs / (min(profile.times) * 1000000)))

      listingFile.write("\n")
      listingFile.write("Instruction Types:\n")
      listingFile.write("==================\n")
      listingFile.write("\n")

      listingFile.write(" %-33s %15s %8s %8s %8s\n" % (
        "Instruction Type", "Executions", "Percent", "Count", "Percent"))
      listingFile.write("\n")

      for type in sorted(InstructionType.getTypes(),
          key=lambda type: profile.getInstExecs(type), reverse=True):

        listingFile.write(" %-33s %15s %7.2f%% %8s %7.2f%%\n" % (
          type.name,
          profile.getInstExecs(type),
          poke.fractionToPercent(
            profile.getInstExecs(type),
            profile.getInstExecs()),
          self.instCountByTypeMap.get(type.name, 0),
          poke.fractionToPercent(
            self.getInstCount(type),
            self.getInstCount())))

      listingFile.write(" %-33s %15s %7.2f%% %8s %7.2f%%\n" % (
        "",
        profile.getInstExecs(),
        100.0,
        self.getInstCount(),
        100.0))


      listingFile.write("\n")
      listingFile.write("Top 10 Functions By Instruction Executions:\n")
      listingFile.write("===========================================\n")
      listingFile.write("\n")

      listingFile.write(" %-49s %8s %8s %8s\n" % (
        "Function Name", "Func %", "BB %", "Inst %"))
      listingFile.write("\n")

      for i, function in enumerate(sorted(
        self.functionMap.values(),
        key=lambda f: sum(profile.blockExecs[b.id] * len(b.instList)
          for b in f.blockList),
        reverse=True)):

        if i == 10:
          break

        name = function.getScopedName()
        if len(name) > 49:
          name = "%s..." % name[:46]

        listingFile.write(" %-49s %7.2f%% %7.2f%% %7.2f%%\n" % (
          name,
          poke.fractionToPercent(profile.functionExecs[function.id],
            profile.totalFunctionExecs),
          poke.fractionToPercent(
            sum(profile.blockExecs[b.id] for b in function.blockList),
            profile.totalBlockExecs),
          poke.fractionToPercent(
            sum(profile.blockExecs[b.id] * len(b.instList)
              for b in function.blockList),
            profile.totalInstExecs)))


      listingFile.write("\n")
      listingFile.write("Top 10 Basic Blocks by Instruction Executions:\n")
      listingFile.write("==============================================\n")
      listingFile.write("\n")

      listingFile.write(" %-49s %8s %8s %8s\n" % (
        "Scoped Basic Block Name", "BB %", "Inst's", "Inst %"))
      listingFile.write("\n")

      for i, block in enumerate(sorted(
        self.blockMap.values(),
        key=lambda b: profile.blockExecs[b.id] * len(b.instList),
        reverse=True)):

        if i == 10:
          break

        name = block.getScopedName()
        if len(name) > 49:
          name = "%s..." % name[:46]

        listingFile.write(" %-49s %7.2f%% %8s %7.2f%%\n" % (
          name,
          poke.fractionToPercent(profile.blockExecs[block.id],
            profile.totalBlockExecs),
          len(block.instList),
          poke.fractionToPercent(
            profile.blockExecs[block.id] * len(block.instList),
            profile.totalInstExecs)))
      listingFile.write("\n")

      listingFile.close()

      # % dynamic instruction executions vs. static instruction count
      dynamicVsStaticFile = open('%s/dynamicVsStatic.txt' % profileDir, 'w')
      profile.writeSubsetsGraph(dynamicVsStaticFile)
      dynamicVsStaticFile.close()

      # instructions
      instListFile = open('%s/instructions.txt' % profileDir, 'w')
      for inst in sorted(
          self.instructionList,
          key=lambda i: profile.blockExecs[i.block.id],
          reverse=True):

        instListFile.write("%-8s %-45s %-15s %-15s %-8x %-8x\n" % (
            inst.id,
            inst.getScopedName(),
            inst.getOpcode().name,
            profile.blockExecs[inst.block.id],
            inst.instrumented,
            inst.relaxer))
      instListFile.close()

      # blocks
      blockListFile = open('%s/blocks.txt' % profileDir, 'w')
      for block in sorted(
          self.blockList,
          key=lambda b: profile.blockExecs[b.id] * len(b.instList),
          reverse=True):

        # relaxer = True
        # for inst in block.instList:
        #   if inst.opcodeName == "Store":
        #     relaxer = False

        blockListFile.write("%-8s %-45s %-8s %-15s %-15s\n" % (
            block.id,
            block.getScopedName(),
            len(block.instList),
            profile.blockExecs[block.id],
            profile.blockExecs[block.id] * len(block.instList)))
      blockListFile.close()

      # functions
      functionListFile = open('%s/functions.txt' % profileDir, 'w')
      functionListFile.write(
          "%-8s %-8s %-45s %-15s %-15s %-15s %-8s %-8s %-15s\n" % (
            "Fn ID",
            "Contr %",
            "Fn Name",
            "Fn Execs",
            "BB Execs",
            "Inst Execs",
            "Store %",
            "Relax %",
            "IPR"))

      for function in sorted(
              self.functionList,
              key=lambda f: sum(profile.blockExecs[b.id] * len(b.instList)
                for b in f.blockList),
              reverse=True):

        instExecs = profile.getInstExecs(function)

        if instExecs != 0:

          instrumentedExecs = 0
          relaxExecs = 0
          storeExecs = 0

          for bb in function.blockList:
            if bb.relax:
              relaxExecs += profile.blockExecs[bb.id]
            for inst in bb.instList:
              instrumentedExecs += inst.instrumented * profile.blockExecs[bb.id]

              if inst.opcodeName == "Store":
                storeExecs += profile.blockExecs[bb.id]

          if relaxExecs:
            IPR = instrumentedExecs / relaxExecs
          else:
            IPR = 0

          functionListFile.write(
            "%-8s %-8.2f %-45s %-15s %-15s %-15s %-8.2f %-8.2f %-15s\n" % (
              function.id,
              poke.fractionToPercent(instExecs, profile.getInstExecs()),
              function.getScopedName(),
              profile.functionExecs[function.id],
              sum(profile.blockExecs[b.id] for b in function.blockList),
              instExecs,
              poke.fractionToPercent(storeExecs, instExecs),
              poke.fractionToPercent(instrumentedExecs, instExecs),
              IPR))
          
      functionListFile.close()

  @staticmethod
  def load(loadDir): #, semantics):
    pickleFile = open("%s/stats.pickle" % loadDir)
    stats = cPickle.load(pickleFile)
    #stats.setSemantics(semantics)
    pickleFile.close()
    return stats

  def getInstCount(self, type=None):
    if type:
      assert isinstance(type, InstructionType)
      return self.instCountByTypeMap[type.name]
    
    return len(self.instructionList)

